/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: cache.c,v 1.22 2003/03/07 17:48:11 thepin Exp $
 */

/*
** The Cache, as the name indicates caches all remote directory content if it is turned on.
** Cache entries expire after certain time defined in junkie preferences.
*/

#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
/* local headers. */
#include <config.h>
#include <cache.h>
#include <ftp.h>
#include <jio.h>
#include <misc.h>
#include <error.h>
#include <mem.h>
#include <defaults.h>

/*
 * 	changed the format of the cache a bit so it will be like
 * 	(perm,owner,group) date size file
 * 	the perm, owner, group are only optional and are
 * 	not always provided, only date, size and file
 * 	can be assumed to exist.
 */
 

#define SOE(x)	((x) ? (x) : (""))			/* self or empty */

static int cache_inited;
static int cache_fd;
static _cache_item *cache_head;


/*
 * Looks up if a site has an entry in the junkie cache, if it does, return
 * the entry, if it does not return NULL.
 */
_cache_item *
find_site(char *site, unsigned long port)
{
	_cache_item *cpnt = NULL;

	cpnt = cache_head;
	while (cpnt) {
		if (!strcmp(cpnt->host, site) && port == cpnt->port)
			return cpnt;
		
		cpnt = cpnt->next;
	}
	
	return NULL;
}

/*
 * Locates a directory (char *directory) for a given site (defined by cpnt)
 * and returns a pointer to a structure containing information concerning that
 * directory.
 */
struct cache_content *
find_content(_cache_item *cpnt, char *directory)
{
	struct cache_content *contpnt = NULL;
	
	if (!cpnt || !directory)
		return NULL;
		
	contpnt = cpnt->content;
	while (contpnt) {
		if (!strcmp(contpnt->directory, directory))
			return contpnt;
		
		contpnt = contpnt->next;
	}

	return NULL;
}

/*
 * Load the cache up from the junkie cache file (configdir/cache),
 * into a linked list.
 */
int
cache_init()
{
	char *buf = NULL, *cacheline = NULL;
	char *date = NULL, *dir = NULL, *pnt = NULL, *host = NULL, *port = NULL;
	char *file = NULL, *mtime = NULL, *perm = NULL, *owner = NULL;
	char *group = NULL, *size = NULL;
	_cache_item **cpnt = NULL, *cpnt2 = NULL, *prior = NULL;
	struct cache_content **site_cont = NULL, *site_cont2 = NULL;
	struct _list **dir_cont = NULL;

	if (getenv("HOME") != 0) {
		buf = j_asprintf("%s/.junkie/cache", getenv("HOME"));
	} else {
		DEBUG_MSG(DBG_WEIRD, "WARNING: $HOME isn't set, trying .");
		buf = j_asprintf("./.junkie/cache");
	}
	
	cache_fd = open(buf, O_RDWR);
	if (cache_fd == -1) {	/* XXX check for permission denied etc ? */
		DEBUG_MSG(DBG_WEIRD, "no cache file found.");
		return 0;
	}
	
	cpnt = &cache_head;
	while ((cacheline = readline(cache_fd)) != NULL) {
		/*
		 * Read a line from the cache file,
		 * parse it and load into memory.
		 */
		if (cacheline[0] == '\t') {
			/*
			 * NEW, format for file entry now: 
			 * "\t(perm,owner,group) [mtime] [size] [file]"
			 */
			/*
			 * XXX FIXME add compability for older versions?
			 */
			
			if (cpnt2 == NULL) {
				/* not belonging to a site */
				/* XXX make more logical var name */ 
				continue;
			}
				

			if ( !(pnt = strchr(cacheline, '(')) ) {
				DEBUG_MSG(DBG_WEIRD, "Invalid cache file,"
				" line not containing (, perhaps an old"
				" version of junkie created this?");
				continue;
			}
			
			perm = pnt + 1;
			if ( !(pnt = strchr(pnt+1, ',')) ) {
				DEBUG_MSG(DBG_WEIRD, "Invalid cache file"
				" (maybe created by old junkie version?)");
				continue;
			}
			*pnt = 0;
			
			if (perm == pnt)
				perm = NULL;
				
			owner = pnt + 1;
			if ( !(pnt = strchr(pnt+1, ',')) ) {
				DEBUG_MSG(DBG_WEIRD,
				"Invalid cache file (maybe created by"
				" an old junkie version?)");
				continue;
			}
			*pnt = 0;
			
			if (owner == pnt)
				owner = NULL;

			group = pnt + 1;
			if ( !(pnt = strchr(pnt+1, ')')) ) {
				DEBUG_MSG(DBG_WEIRD,
				"Invalid cache file (maybe created by"
				" an old junkie version?)");
				continue;
			}
			*pnt = 0;
			
			if (group == pnt)
				group = NULL;
			
					
			mtime = pnt + 2;
			if ( (pnt = strchr(pnt+2, ' ')) == NULL)
				 continue;
			*pnt = 0;

			size = pnt + 1;
			if ( (pnt = strchr(size, ' ')) == NULL)
				continue;
			*pnt = 0;

			file = pnt + 1;
			if ( (*dir_cont = (struct _list *)
			j_malloc(sizeof(struct _list))) == NULL) 
				return -1;
			
			(*dir_cont)->file = (*dir_cont)->perm = NULL;
			(*dir_cont)->owner = (*dir_cont)->group = NULL;
			(*dir_cont)->next = NULL;
			
			(*dir_cont)->file = j_strdup(file);
			(*dir_cont)->perm = j_strdup(perm);
			(*dir_cont)->owner = j_strdup(owner);
			(*dir_cont)->group = j_strdup(group);
				
			(*dir_cont)->mtime = atol(mtime);
			(*dir_cont)->size = atol(size);
			
			switch (perm[0]) {
				case 'd':
					(*dir_cont)->type = _LIST_DIR;
					break;
				case '-':
					(*dir_cont)->type = _LIST_FILE;
					break;
				case 'l':
					(*dir_cont)->type =
						_LIST_DIR | _LIST_FILE;
					break;
			}
			dir_cont = &(*dir_cont)->next;
			 
			continue;
		}
		 
		/* [date] [site:port] [directory] */
		date = cacheline;
		if ( (pnt = strchr(cacheline, ' ')) == NULL)
			continue;
		*pnt = 0;

		host = pnt + 1;
		if ( (pnt = strchr(host, ' ')) == NULL)
			continue;
		*pnt = 0;

		dir = pnt + 1;
		if ( (pnt = strchr(host, ':')) == NULL)
			continue;
		*pnt = 0;
		port = pnt + 1;
		
		/* done setting all variables */
		if ( (cpnt2 = find_site(host, atol(port))) == NULL) {
			if ( (*cpnt = (_cache_item *)
			j_malloc(sizeof(_cache_item))) == NULL) 
				return -1;
			
			(*cpnt)->prior = prior;
			(*cpnt)->next = NULL;
			(*cpnt)->content = NULL;
		
			site_cont = &(*cpnt)->content;
			if ( ((*cpnt)->host = strdup(host)) == NULL) {
				MEM_ERR("strdup", ERR_SHOW);
				return -1;
			}
		
			(*cpnt)->port = atol(port);
			cpnt2 = *cpnt;
			prior = *cpnt;
			cpnt = &(*cpnt)->next;
		} else {
			site_cont = &cpnt2->content;
			while (*site_cont)
				site_cont = &(*site_cont)->next;
		}

		if ( (site_cont2 = find_content(cpnt2, dir)) == NULL) {
			if ( (*site_cont = (struct cache_content *)
			j_malloc(sizeof(struct cache_content))) == NULL) 
				return -1;

			(*site_cont)->next = NULL;
			(*site_cont)->files = NULL;
		
			(*site_cont)->directory = j_strdup(dir);
		 	(*site_cont)->date = atol(date);

			dir_cont = &(*site_cont)->files;
			site_cont2 = *site_cont;
			site_cont = &(*site_cont)->next;
		} else {
			dir_cont = &site_cont2->files;
		}
		
		j_free(cacheline);
	}
	
	j_free(buf);
	cache_inited = 1;

	return 0;
}

/*
 * Write the cache (loaded in memory) down to the junkie cache file.
 */
int
cache_save()
{
	_cache_item *cpnt = NULL;
	struct cache_content *dpnt = NULL;
	struct _list *fpnt = NULL;
	char *buf = NULL, *bracket = NULL;
	jio *cache_fh = NULL;

	if (getenv("HOME") != 0) {
		buf = j_asprintf("%s/.junkie/cache", getenv("HOME"));
	} else {
		DEBUG_MSG(DBG_WEIRD, "WARNING: $HOME isn't set, trying .");
		buf = j_asprintf("./.junkie/cache");
	}
	
	cache_fh = j_open(buf, O_RDWR | O_CREAT | O_TRUNC, 0600);
	if (!cache_fh) {
		DEBUG_MSG(DBG_ERROR, "failed to create cache file %s", buf);
		SYS_ERR("open", ERR_SHOW);
		cache_inited = -1;
		return -1;
	}
	j_free(buf);

	/* write the shit. */
	cpnt = cache_head;
	while (cpnt) {
		dpnt = cpnt->content;
		while (dpnt) {
			/*
			 * Write out the information of a site 
			 * and a directory in the cache
			 */
			buf = j_asprintf("%d %s:%d %s\n",
				(int)dpnt->date, cpnt->host, (int)cpnt->port,
				dpnt->directory);
			
			j_write(cache_fh, buf, strlen(buf));
			j_free(buf);
			
			/*
			 * For that directory, go through all content 
			 * of that dir and write it 
			 */
			fpnt = dpnt->files;
			while (fpnt) {
				if (fpnt && fpnt->perm && fpnt->owner 
				&& fpnt->group && fpnt->file) {
					bracket = j_asprintf("(%s,%s,%s)",
		 					SOE(fpnt->perm), 
							SOE(fpnt->owner),
							SOE(fpnt->group));
					
					buf = j_asprintf("\t%s %d %d %s\n",
							bracket, 
							(int)fpnt->mtime,
							(int)fpnt->size,
							fpnt->file);
					
					j_write(cache_fh, buf, strlen(buf));
					
					j_free(buf);
					j_free(bracket);
				}
				
				fpnt = fpnt->next;
			}
			dpnt = dpnt->next;
		}
		cpnt = cpnt->next;
	}
	
	j_close(cache_fh);
	
	return 0;
}

/*
 * add/update entry, for example when the user does Refresh,
 * or when there is no data available in the cache.
 */
int
cache_update(_ftp_item *site, char *dir, struct _list *files)
{
	_cache_item **cpnt = NULL, *cpnt2 = NULL, *prior = NULL;
	struct cache_content **dpnt = NULL, *dpnt2 = NULL, *dpnt3 = NULL;
	struct _list **fpnt = NULL;

	if (site == NULL || dir == NULL || files == NULL)
		return -1;

	/* initialise cpnt */
	cpnt = &cache_head;
	if (*cpnt == NULL) {
		if ( ((*cpnt) = (_cache_item *)
		j_malloc(sizeof(_cache_item))) == NULL) 
			return -1;
	
		(*cpnt)->host = strdup(site->addr);
		(*cpnt)->port = site->port;
		
		(*cpnt)->content = NULL;
		(*cpnt)->next = NULL;
		(*cpnt)->prior = NULL;
	} else {
		if ( (cpnt2 = find_site(site->addr, site->port)) == NULL) {
			while (*cpnt) {
				prior = *cpnt;
				cpnt = &(*cpnt)->next;
			}

			if ( ((*cpnt) = (_cache_item *)
			j_malloc(sizeof(_cache_item))) == NULL) 
				return -1;
			
			(*cpnt)->host = j_strdup(site->addr);
			(*cpnt)->port = site->port;

			(*cpnt)->content = NULL;
			(*cpnt)->prior = prior;
			(*cpnt)->next = NULL;
		} else {
			cpnt = &cpnt2;
		} 
	}

	/* continue */
	if ( (dpnt2 = find_content((*cpnt), dir)) != NULL) {
		if (dpnt2 == (*cpnt)->content) {
			(*cpnt)->content = (*cpnt)->content->next;
		} else {
			dpnt3 = (*cpnt)->content;
			while (dpnt3) {
				if (dpnt3->next == dpnt2)
					break;
				dpnt3 = dpnt3->next;
			}

			if (dpnt3 != NULL) {
				dpnt3->next = dpnt2->next;
			}
		}

		cache_free_content(dpnt2);
	}
	
	dpnt = &(*cpnt)->content;
	while (*dpnt) dpnt = &(*dpnt)->next;

	if ( ((*dpnt) = (struct cache_content *)
	j_malloc(sizeof(struct cache_content))) == NULL) 
		return -1;
	
	(*dpnt)->directory = j_strdup(dir);
	if ((*dpnt)->directory == NULL) {
		MEM_ERR("strdup", ERR_SHOW);
		return -1;
	}
	(*dpnt)->date = time(NULL);
	(*dpnt)->files = NULL;
	(*dpnt)->next = NULL;
	
	fpnt = &(*dpnt)->files;
	*fpnt = 0;

	/* add files from struct _list *files */
	while (files) {
		if ( !((*fpnt) = (struct _list *)
		j_malloc(sizeof(struct _list))) ) 
			return -1;

		(*fpnt)->file = (*fpnt)->perm = NULL;
		(*fpnt)->owner = (*fpnt)->group = NULL;
		(*fpnt)->mtime = (*fpnt)->size = (*fpnt)->type = 0;
		
		(*fpnt)->file = j_strdup(files->file);
		(*fpnt)->perm = j_strdup(files->perm);
		(*fpnt)->owner = j_strdup(files->owner);
		(*fpnt)->group = j_strdup(files->group);
		
		(*fpnt)->size = files->size;
		(*fpnt)->mtime = files->mtime;
		(*fpnt)->type = files->type;
		(*fpnt)->next = NULL;

		fpnt = &(*fpnt)->next;
		files = files->next;
	}

	return 1;
}

/*
 * returns a list of files or 0 if the cache entry is not avail.
 */
struct _list *
cache_list(_ftp_item *ftp, char *dir)
{
	_cache_item *cpnt = NULL;
	struct cache_content *dpnt = NULL;
	struct _list *list_dir = NULL, **lpnt = NULL;
	struct _list *list_ent = NULL;
	char *expdate = NULL;
	int exp;

	if ( (cpnt = find_site(ftp->addr, ftp->port)) == NULL)
		return NULL;
	
	if ( (dpnt = find_content(cpnt, dir)) == NULL)
		return NULL;

	/*  has the cache entry expired? */ 
	if (config_read_value("cache-days", T_NUM, 
	(void **)&expdate) == 0 && expdate) {
		/* seconds = days * 24 * 3660 */
		exp = atol(expdate) * 24 * 3660;
		if (exp > 0) {
			DEBUG_MSG(DBG_INFO, 
			     "cache entry expires after %d seconds", 
			     (int)((dpnt->date + exp) - time(NULL)));
			
			if ((dpnt->date + exp) < time(NULL)) {
				DEBUG_MSG(DBG_INFO, 
				"cache entry has already expired");
				return NULL;
			}
		}
	}

	list_ent = dpnt->files;
	lpnt = &list_dir;
	while (list_ent) {
		if ( ((*lpnt) = (struct _list *)
		j_malloc(sizeof(struct _list))) == NULL) 
			return NULL;

		memset(*lpnt, 0, sizeof(struct _list));
		(*lpnt)->file = j_strdup(list_ent->file);
		(*lpnt)->perm = j_strdup(list_ent->perm);
		(*lpnt)->owner = j_strdup(list_ent->owner);
		(*lpnt)->group = j_strdup(list_ent->group);

		(*lpnt)->size = list_ent->size;
		(*lpnt)->mtime = list_ent->mtime;
		(*lpnt)->type = list_ent->type;
		(*lpnt)->next = NULL;
		lpnt = &(*lpnt)->next;
		list_ent = list_ent->next;
	}

	return list_dir;
}


void
cache_free_dir(struct _list *cdc)
{
	struct _list *cdc2 = NULL;

	while (cdc) {
		cdc2 = cdc;
		cdc = cdc->next;
		j_free(cdc2->file);
		j_free(cdc2->perm);
		j_free(cdc2->owner);
		j_free(cdc2->group);
		cdc2->file = cdc2->perm = NULL;
		cdc2->owner = cdc2->group = NULL;
		cdc2->next = NULL;
		j_free(cdc2);
	}

	return;
}

void
cache_free_content(struct cache_content *cc)
{
	cache_free_dir(cc->files);
	cc->files = NULL;
	j_free(cc->directory);
	j_free(cc);

	return;
}


void
cache_free_site(_ftp_item *ftp)
{
	_cache_item *cc = NULL;
	struct cache_content *contpnt = NULL, *contpnt2 = NULL;

	if (!ftp)
		return;

	if ( (cc  = find_site(ftp->addr, ftp->port)) == NULL)
		return;

	contpnt = cc->content;
	while (contpnt) {
		contpnt2 = contpnt;
		contpnt = contpnt->next;
		cache_free_content(contpnt2);
	}
	
	if (cc->prior)
		cc->prior->next = cc->next;
	else {
		if (cc->next)
			cc->next->prior = NULL;
			
		cache_head = cache_head->next;
	}

	j_free(cc);

	return;
}
