/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: ftp_net.c,v 1.34 2003/03/03 02:16:33 tgucio Exp $
 */


#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/stat.h>
/* local headers */
#include <ftp.h>
#include <ftp_misc.h>
#include <ftp_net.h>
#include <ftp_event.h>
#include <ftp_request.h>
#include <ftp_resolver.h>
#include <hammering.h>
#include <error.h>
#include <medium.h>
#include <telnet.h>
#include <log.h>
#include <mem.h>
#include <jfork.h>
#include <jlist.h>
#include <misc.h>
#include <defaults.h>
#include <gui_toolkit.h>	/* XXX :/ */

void
connect_to_lookup_cb(jslist *jlist, void *param)
{
	ftp_resolver_info *rinfo = NULL;
	ftp_object *obj = NULL;
	jslist *jnode = NULL;
	cb callback;
	int port;
	void *data = NULL;
	int fd;

	DEBUG_MSG(DBG_INFO, "ftp_connect_lookup_cb");

	/* get the various arguments */
	obj = param;
	callback = ftp_object_get_data(&obj, "connect_to_callback");
	data = ftp_object_get_data(&obj, "connect_to_data");
	port = (int)ftp_object_get_data(&obj, "connect_to_port");
	ftp_object_free(&obj); obj = NULL;
	
	
	/* if there is no callback, there is no purpose of continuing.*/
	if (!callback)  {
		DEBUG_MSG(DBG_WEIRD, "no callback");
		return;
	}

	if (!jlist) {
		DEBUG_MSG(DBG_ERROR, "no jlist");
		return;
	}

	jnode = jlist;
	while (jnode) {	
		rinfo = (ftp_resolver_info *)jnode->data;

		if ( (fd = socket(rinfo->ai_family, SOCK_STREAM, rinfo->ai_proto)) < 0) {
			SYS_ERR("socket", ERR_SHOW);
			continue;
		}

		if (fcntl(fd, F_SETFL, O_NONBLOCK) < 0) {
			callback(-1, 1, data);
			break;
		}

		// set the port
		ftp_resolver_info_set_port(rinfo, port);
		
		if (connect(fd, rinfo->addr, rinfo->addr_len) == 0
		   || errno == EINPROGRESS || errno == EAGAIN)  {
			/* connected. */
			callback(fd, 0, data);
			break;
		}
		close(fd);
		
		jnode = jnode->next;
	}
	
	if (!jnode)
		callback(-1, 1, data);
	
	return;
}


/*
 * "non-blocking" connect to a host,
 * calls the provided callback when connected.
 */
int
connect_to(char *host, int port, cb callback, void *data)
{
	ftp_object *obj = NULL;

	if (!host || port <= 0)
		return 1;
		
	ftp_object_set_data(&obj, "connect_to_callback", callback);
	ftp_object_set_data(&obj, "connect_to_data", data);
	ftp_object_set_data(&obj, "connect_to_port", (void *)port);
	ftp_resolver_lookup(host, connect_to_lookup_cb, obj);	
	
	return 0;
}


int
bind_to(int port)
{
	struct sockaddr_in serv_addr;
	char reuse = 1;
	int sfd;

	if ( (sfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		SYS_ERR("socket", ERR_SHOW);
		return -1;
	}

	if(fcntl(sfd, F_SETFL, O_NONBLOCK) < 0)
		return -1;

	memset(&serv_addr, 0, sizeof(struct sockaddr_in));
	
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);
	serv_addr.sin_addr.s_addr = INADDR_ANY;

	if (setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) == 0) {
		SYS_ERR("setsockopt", ERR_SHOW);
		return -1;
	}

	if (bind(sfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) == -1) {
		SYS_ERR("bind", ERR_SHOW);
		return -1;
	}

	if (listen(sfd, 1) == -1 && errno != EINPROGRESS) {
		SYS_ERR("listen", ERR_SHOW);
		return -1;
	}
	
	return sfd;
}


#ifdef IPV6
int bind6_to (int port, int af) {
	int sfd,reuse=1;
	struct sockaddr_in6 sin6;
	
	sin6.sin6_addr=in6addr_any;
	sin6.sin6_family=af;
	sin6.sin6_port=htons(port);
	sin6.sin6_flowinfo=0;
	
	if ((sfd = socket(af, SOCK_STREAM, IPPROTO_TCP)) == -1) {
		SYS_ERR("socket", ERR_SHOW);
		return -1;
	}

	if(fcntl(sfd, F_SETFL, O_NONBLOCK) < 0)
		return -1;

	if (setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) != 0) {
		SYS_ERR("setsockopt", ERR_SHOW);
		return -1;
	}
	
	if (bind(sfd, (struct sockaddr*)&sin6, sizeof(sin6)) != 0) {
		SYS_ERR("bind", ERR_SHOW);
		return -1;
	}

	if (listen(sfd, 1) == -1 && errno != EINPROGRESS) {
		SYS_ERR("listen", ERR_SHOW);
		return -1;
	}

	return sfd;
}
#endif

int
accept_dataport(int sfd)
{
	int fd;
#ifdef IPV6
	int size = sizeof(struct sockaddr_in6);
	struct sockaddr_in6 sin;
#else
	int size = sizeof(struct sockaddr_in);
	struct sockaddr_in sin;
#endif

	if ( (fd = accept(sfd, (struct sockaddr *)&sin, (socklen_t *)&size)) == -1) {
		DEBUG_MSG(DBG_ERROR, "accept, errno: %d", errno);
		SYS_ERR("accept", ERR_SHOW);
		return -1;
	}

	return fd;
}

/*
** Sends a command to the ftp site (possible to use a format string).
** Also logs the command in the log window.
*/
int
ftp_send_command(_ftp_item *ftp, char *command, ...)
{
	va_list val;
	char buf[GOOD_SIZE + 1];
	memset(buf, 0, sizeof(buf));

	if (ftp == NULL)
		return -1;

	/* parse variable argument list. */
	va_start(val, command);
	vsnprintf(buf, GOOD_SIZE, command, val);
	va_end(val);

	if (strncasecmp(buf, "PASS ", 5)) {
		FTP_SENT_LOG(ftp, "    %s\n", buf);
	} else {
		FTP_SENT_LOG(ftp, "    PASS xxxxxxxx\n");
	}


	/* send the command and return its response. */
	return (send(ftp->fd, buf, strlen(buf), 0));
}



/*
** Reads in one line (a sequence of characters followed by a newline (\n)).
** through the ftp_read_char returns it through the /buf/ user-argument.
** Returns 0 for no connection or was connection closed, -1 on error,
** or normally (when successful) the number of bytes read.
*/
int
ftp_readline(_ftp_item *ftp, unsigned char *line, int buffer_size, int readmore)
{
	/* non-static shit. */
	char *pnt = NULL, *ptr = NULL;
	char *pnt2 = NULL;
	unsigned long size;
	int len;

	if (!ftp || !line) 
		return FTP_READ_FAILURE;

	/* check if there is a cached line available. */
	if (ftp->bf.buffer && ftp->bf.buffer_cached > 0) {
		/* if a whole line is available return it. */
		if ((pnt = strchr(ftp->bf.buffer, '\n'))) {
			*pnt = 0;
			if ((pnt2 = strchr(ftp->bf.buffer, '\r')))
				*pnt2 = 0;

			strncpy(line, ftp->bf.buffer, buffer_size);
			
			ftp->bf.buffer_cached = strlen(pnt+1);
	
			/* remove the returned line from buf. */
			memcpy(ftp->bf.buffer, pnt+1, ftp->bf.buffer_cached);
			ftp->bf.buffer[ftp->bf.buffer_cached] = 0;

			/* resize */
			if (!(ptr = realloc(ftp->bf.buffer, ftp->bf.buffer_cached + 1))) {
				free(ftp->bf.buffer);
				return FTP_READ_FAILURE; 
			}

			ftp->bf.buffer = ptr;

			return FTP_READ_SUCCESS;
		}
	}

	if (!readmore)
		return FTP_READ_CACHED;
		
	/* no cached line available, read more. */
	size = ftp->bf.buffer_cached + MAX_SIZE;

	if (!(ptr = realloc(ftp->bf.buffer, size + 1))) {
		free(ftp->bf.buffer);
		return FTP_READ_FAILURE;
	}
	ftp->bf.buffer = ptr;
	
	if ((len = read(ftp->fd, ftp->bf.buffer + ftp->bf.buffer_cached, MAX_SIZE)) <= 0) {
		if (len == 0)
			return FTP_CONN_CLOSED;

		return FTP_CONN_ERR;
	}

	*(ftp->bf.buffer + ftp->bf.buffer_cached + len) = 0;
	ftp->bf.buffer_cached = size;

	/* if more space is allocated than needed then resize. */
	if (strlen(ftp->bf.buffer) != ftp->bf.buffer_cached) {
		ftp->bf.buffer_cached = strlen(ftp->bf.buffer);
		if (!(ptr=realloc(ftp->bf.buffer, ftp->bf.buffer_cached + 1))) {
			free(ftp->bf.buffer);
			return FTP_READ_FAILURE;
		}
		ftp->bf.buffer = ptr;
	}

	/* if there is a whole line available return it, otherwise return NULL */
	if ((pnt = strchr(ftp->bf.buffer, '\n'))) {
		*pnt = 0;
		if ((pnt2 = strchr(ftp->bf.buffer, '\r')))
			*pnt2 = 0;
		strncpy(line, ftp->bf.buffer, buffer_size);
		ftp->bf.buffer_cached = strlen(pnt+1);
	
		/* remove the returned line from buf. */
		memcpy(ftp->bf.buffer, pnt+1, ftp->bf.buffer_cached);
		ftp->bf.buffer[ftp->bf.buffer_cached] = 0;

		/* resize */
		if (!(ptr = realloc(ftp->bf.buffer, ftp->bf.buffer_cached + 1))) {
			free(ftp->bf.buffer);
			return FTP_READ_FAILURE;
		}

		ftp->bf.buffer = ptr;

		return FTP_READ_SUCCESS;
	}

	return FTP_READ_CACHED;
}


/*
** Reads in one response from an FTP server and returns it through the user-argument buf.
** On error, it will return the exact error return-value given by ftp_readline,
** and when successful, will return the FTP code of the response from the FTP server.
*/
int
ftp_get_resp(_ftp_item *ftp, unsigned char *buf, int buf_size)
{
	int code = 0, code2 = 0, ret, readmore = 1;

	if (ftp == NULL)
		return 0;

	code = (int)ftp_object_get_data(FTP_OBJECT(ftp), "ftp_code");
	code2 = (int)ftp_object_get_data(FTP_OBJECT(ftp), "ftp_code2");
	while (1) {
		memset(buf, 0, buf_size);
		if ((ret = ftp_readline(ftp, buf, buf_size, readmore)) <= 0)
			return ret;
			
		if (readmore)
			readmore = 0;

		switch (ret) {
		case FTP_CONN_CLOSED:
			return 0;
			break;
		case FTP_CONN_ERR:
		case FTP_READ_FAILURE:
			return -1;
			break;
		case FTP_READ_CACHED:
			ftp_object_set_data(FTP_OBJECT(ftp), "ftp_code", (void *)code);
			ftp_object_set_data(FTP_OBJECT(ftp), "ftp_code2", (void *)code2);
			return FTP_READ_CACHED;
		}

		/* Log received responses. */
		FTP_RECV_LOG(ftp, "    %s\n", buf);

		if (code == 0 && buf[3] == ' ')
			break;

		code2 = (buf[0] - '0')*100 + (buf[1] - '0')*10 + (buf[2] - '0');
		if (code == 0 && buf[3] == '-') 
			code = code2;
		else if (buf[3] == ' ' && code == code2)
			break;

	}

	ftp_object_unlink_tag(FTP_OBJECT(ftp), "ftp_code");
	ftp_object_unlink_tag(FTP_OBJECT(ftp), "ftp_code2");
	code = (buf[0] - '0')*100 + (buf[1] - '0')*10 + (buf[2] - '0');

	return code;
}



/*
** Only connects to an FTP server and returns the fd.
** Does nothing more than that.
XXX kill?
int
ftp_open(_ftp_item *ftp, cb callback, void *data)
{
	return ftp_connect_to(ftp, callback, data);
}
*/


/*
** callback
*/
void
ftp_connect_cb(_ftp_item *ftp, int result, void *notused)
{
	cb callback;
	void *data = NULL;
	
	DEBUG_MSG(DBG_INFO, "ftp_connect_cb");

	callback = ftp_object_get_data(FTP_OBJECT(ftp), "ftp_connect_callback");
	data = ftp_object_get_data(FTP_OBJECT(ftp), "ftp_connect_data");

	DEBUG_MSG(DBG_INFO, "ftp_connect_cb");
	if (result || !callback) {
		DEBUG_MSG(DBG_INFO, "failed, result: %d", result);
		j_free(ftp);
		callback(NULL, data);
		return;	/* failed to connect. */
	}
	
	/* setup event handlers. */
	DEBUG_MSG(DBG_INFO, "setting up some sorta stuff");
	ftp_event_setup_default(ftp);
	hammering_enable_handlers(ftp);
	callback(ftp, data);

	DEBUG_MSG(DBG_INFO, "go00: ftp->fd: %d", ftp->fd);
	return;
}

/*
** Normal connect, will set the default handlers etc.
** Users wanting different event handlers etc should make up their own connect function
** using ftp_open etc..
** This function will allocate a new ftp_item entry.
*/
_ftp_item *
ftp_connect(const char *host, int port, int mode, const char *user, const char *pass, cb callback, void *data)
{
	_ftp_item *ftp = NULL;

	if (!host || port <= 0 || !user || !pass)
		return NULL;
	
	/* allocate an entry. */
	if (!(ftp = ftp_new()))
		return NULL;

	/* set the information for the ftp_item. */
	ftp_set_type(ftp, CONN_FTP);
	ftp_set_server(ftp, host, port);
	ftp_set_user(ftp, user, pass);
	ftp_set_mode(ftp, mode);			/* set some mode.  PASV/PORT */ 
	ftp->requests = 0;

	// set info
	ftp_object_set_data(FTP_OBJECT(ftp), "ftp_connect_callback", callback);
	ftp_object_set_data(FTP_OBJECT(ftp), "ftp_connect_data", data);

	/* connect to the ftp :) */
	if (!ftp_reconnect(ftp, ftp_connect_cb, NULL)) 
		return NULL;	/* failed. */

	return ftp;	
}


/*
** callback for ftp_reconnect
*/
void
ftp_reconnect_cb(int fd, int ret, void *param)
{
	_ftp_item *ftp = NULL;
	struct sockaddr_in sin;
	size_t p = sizeof(struct sockaddr_in);
	void *data = NULL;
	cb callback;
	int x;

	/* get arguments */
	ftp = param;
	
	callback = ftp_object_get_data(FTP_OBJECT(ftp), "ftp_reconnect_callback");
	ftp_object_unlink_tag(FTP_OBJECT(ftp), "ftp_reconnect_callback");

	data = ftp_object_get_data(FTP_OBJECT(ftp), "ftp_reconnect_data");
	ftp_object_unlink_tag(FTP_OBJECT(ftp), "ftp_reconnect_data");

	
	DEBUG_MSG(DBG_INFO, "ftp_reconnect_cb");
	if (!ftp || ret || !callback || fd <= 0) {
		DEBUG_MSG(DBG_ERROR, "!ftp || ret || !callback || fd <= 0");
		return;
	}

	ftp->fd = fd;
	ftp->state = FTP_STATE_CONNECTED;
		
	DEBUG_MSG(DBG_INFO, "after ftp_open, fd: %d", ftp->fd);
	/* getpeername should be enough... not needed? */
	if (medium_cmd(M_SELECT_WRITE, &x, &ftp->fd)) {
		if (x) {
			DEBUG_MSG(DBG_ERROR, "medium_cmd failed");
			return;
		}
	}
	
	DEBUG_MSG(DBG_INFO, "after medium_cmd(gui_io_select_write)");
	
	/* getpeername will verify if the connection was opened or not. */
	if (getpeername(ftp->fd, (struct sockaddr *)&sin, &p) != 0) {
		DEBUG_MSG(DBG_ERROR, "getpeername(ftp->fd)... failed");
//		gui_connect_button_set_connect();
		callback(ftp, 1, data);
		return;
	}

	DEBUG_MSG(DBG_INFO, "succes, call bacl");
		
	/* success, call back */
	callback(ftp, 0, data);
	
	return;
}

/*
** reconnect using an existing ftp_item
*/
int
ftp_reconnect(_ftp_item *ftp, cb callback, void *data)
{
/*	struct sockaddr_in sin;
	int p = sizeof(struct sockaddr_in);*/
/*	int ret;*/
	
	if (!ftp)
		return 0;

	/* free the old requests.. etc. from the old session, if there are any. :) */
	ftp->n_requests = 0;
	if (ftp->requests) 
		ftp_request_tree_free(&ftp->requests);

	memset(&ftp->bf, 0, sizeof(struct buffering));

	// set info.
	ftp_object_set_data(FTP_OBJECT(ftp), "ftp_reconnect_callback", callback);
	ftp_object_set_data(FTP_OBJECT(ftp), "ftp_reconnect_data", data);

	/* open the new connection. */
	if (connect_to(ftp->addr, ftp->port, ftp_reconnect_cb, ftp) == -1) {
		DEBUG_MSG(DBG_ERROR, "connect_to() failed");
		goto connection_failure;
	}

	return 1;
connection_failure:
	FTP_MISC_LOG(ftp, "    Failed to connect to %s: %d\n", ftp->addr, ftp->port);
	return 0;
}


int
ftp_close(_ftp_item *ftp)
{
	if (ftp->fd)
		close(ftp->fd);
	
	ftp->fd = -1;
	
	if (ftp->dataconn && ftp->dataconn->fd)
		close(ftp->dataconn->fd), ftp->dataconn->fd = -1;

	return 0;
}

/*
 * ftp_main_close disconnects any existing socket connections
 * and free's the struct _list, provided it is non-NULL
 */
void
ftp_main_close(int fd, struct _list *f)
{
	if(fd > 0)
		close(fd);

	fd = -1;

	if(f)
		ftp_free_list(f);
}

