/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: ftp_request.c,v 1.29 2003/03/07 17:48:12 thepin Exp $
 */


#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/stat.h>
/*temporary*/
#include <gui_ftp.h>

/* local headers */
#include <ftp.h>
#include <ftp_request.h>
#include <ftp_net.h>
#include <defaults.h>
#include <error.h>
#include <log.h>
#include <local.h>
#include <mem.h>

/*
 * ftp_request_find_entry(root, condition, stop_condition)
 * look for a request in a request hierarchy by it's condition,
 * return the first entry that fulfills the condition
 * the stop condition makes it stop searching when it encounters
 * a request that matches that state.
 */
ftp_request *
ftp_request_find_entry(ftp_request *root, int condition, int stop_condition)
{
	ftp_request *rpnt = NULL, *rtree = NULL;

	/* recurses through the tree. */
	if (root == NULL) 
		return NULL;

	rpnt = root;
	while (rpnt) {
		if (stop_condition > 0 && (rpnt->state & stop_condition))
			goto not_found;
			
		if (rpnt->state & condition)
			return rpnt;

		/* go through the subrequests. */
		rtree = rpnt->tree_next;
		while (rtree) {
			if (stop_condition > 0 && (rtree->state & stop_condition))
				goto not_found;
				
			if (rtree->state & condition)
				return rtree;
				
			rtree = rtree->tree_next;
		}

		rpnt = rpnt->next;
	}

not_found:
	return NULL;
}


/*
** Free the members of a single request entry.
*/
void
ftp_request_free(ftp_request *req)
{
	if (!req)
		return;

	/* free the stuff. */
	j_free(req->cmd);
	j_free(req->args);
	ftp_object_free(FTP_OBJECT(req));
	j_free(req);

	return;
}


/*
** Free a tree of request and NULL the entry.
*/
void
ftp_request_tree_free(ftp_request **req)
{
	ftp_request *pnt = NULL, *pnt2 = NULL, *rtree = NULL, *rtree2 = NULL;

	if (req == NULL)
		return;

	pnt = *req;
	while (pnt) {
		rtree = pnt->tree_next;
		while (rtree) {
			rtree2 = rtree;
			rtree = rtree->tree_next;
			ftp_request_free(rtree2);
		}

		pnt2 = pnt;
		pnt = pnt->next;
		ftp_request_free(pnt2);
	}

	*req = NULL;
	return;
}


/*
** Used when debugging the tree.  Will recurse through the tree, printing out information on all members
** of the tree.
*/
ftp_request *
ftp_request_tree_debug(ftp_request *root, int l)
{
	ftp_request *rpnt = NULL;
	
	if (root == NULL)
		return NULL;

	/* Fix this function better. */
	rpnt = root;
	while (rpnt) {
		int i;
		for (i = 0; i < l; i++)
			printf("    ");

		printf("%s %s (status: %d) (req: %p obj: %p)\n", rpnt->cmd, (rpnt->args ? rpnt->args : ""), rpnt->state, rpnt, FTP_OBJECT(rpnt));
		ftp_request_tree_debug(rpnt->tree_next, l + 1);

		rpnt = rpnt->next;
	}

	return NULL;
}

/*
** This function handles checking if there are any requests to be carried out, and if there are,
** it will carry them out.
*/
int
ftp_request_iteration(_ftp_item *ftp)
{
	ftp_request *rpnt = NULL;


	if (!ftp) 
		return 0;

	if (ftp->state == FTP_STATE_NONE || ftp->state == FTP_STATE_CONNECTED)
		return 0;
		
	if (ftp->requests == 0 || ftp->n_requests <= 0) {
		gui_ftp_setup_write(ftp, 0);
		return 0;
	}

	/*
	** Look for a queued entry marked by R_QUEUED but stopping on any requests marked by R_SENT or R_READY.
	** ftp->requests is the root of all requests for that given ftp_site.
	** Will find the next request to be "carried out"/sent yet waiting for the request sent before to be completed (response received).
	*/
	rpnt = ftp_request_find_entry(ftp->requests, R_QUEUED, R_SENT | R_READY);
	if (rpnt == NULL) {
		/*
		** no request was found, turning the request mode off so this function will not be called till a request has been added
		** through ftp_request_add.
		*/
		gui_ftp_setup_write(ftp, 0);
		return 0;
	}

	/*
	** If the request to be issued is a non-subrequest (a main request) then clear
	** all requests prior to it.  XXX (check if this is ok).
	*/
//	if (rpnt->tree_prior == NULL) {
//		ftp_request_tree_free_backwards(rpnt->prior);
//		rpnt->prior = NULL;
//		ftp->requests = rpnt;
//	} 

#ifdef DEBUG
	/*
	** DEBUG MODE.  Print out all information on the request tree.
	** Useful while debugging the ftp code. 
	*/
	printf("DEBUG START________________________________----\n");
	// ftp_request_tree_debug(ftp->requests, 0);
#endif

	/*
	** There are two types of requests,
	**	the first one is REQ_SEND:
	**		This type of requst is used for sending out commands directly to the FTP server and will then, through the event support
	**		call a callback when the response has been received, if a callback was specified.
	**	the second one would be REQ_CMD:
	**		This will simply call the callback specified, the only reason why it might be useful is that it will be called at a time
	**		when other requests that were present in the request list at the time it was added, will already have been processed,
	**		sent out and response received (for REC_SEND requests at least).  Therefore the only usefulness of this command is its
	**		"timing", like if u want to get a folder list and then call a function to do something with the folder list, process it in
	**		some way, then you would send the REQ_SEND message with "LIST..." and then add the processing function as a  REQ_CMD request
	**		with the processing function as a callback.  Then the processing function would be called after the FTP listing had been
	**		finished.
	*/
	if (rpnt->type == REQ_SEND && rpnt->state == R_QUEUED) {
		/*
		** The REQ_SEND type of requests is to be handled here.
		*/
		if(*rpnt->cmd) {
			/*
			** Send the ftp request and mark the request as having been sent.
			*/
			ftp_send_command(ftp, "%s%s%s\r\n", 
				rpnt->cmd, rpnt->args ? " " : "", rpnt->args ? rpnt->args : "");
			
			rpnt->state = R_SENT;
			ftp->n_requests--;		/* decrease the number of requests that need to be sent out. */
			ftp->last = time(NULL);		/* reset the time of the last request sent out. */
			
			/* this will make sure this __FUNCTION__, won't be called till the response to this request  has been received.  */
			gui_ftp_setup_write(ftp, 0);
		}
		else if (rpnt->callback) {
			/*
			** No command to send out was available.  However a callback was available, so the callback will be called
			** here and the state set to SUCCESS though there is no way of knowing if there was a "real success" or not.
			*/
			rpnt->callback(ftp, 0, R_SUCCESS, NULL, NULL, NULL, FTP_OBJECT(rpnt));
			rpnt->state = R_SUCCESS;
			ftp->n_requests--;		/* decrease the number of requests awaiting to be sent out. */
		}		
	} else if (rpnt->type == REQ_CMD && rpnt->state == R_QUEUED) {
		/*
		** The REQ_CMD type, which is only for calling a shipped callback is handled below.  The main/only advantage of the
		** REQ_CMD as described above, is its "timing", that it will be called after requests above in the request list
		** at the time when the request is added will have been processed before the callback for the REQ_CMD is called.
		*/
		if (rpnt->callback) { 
			/* the callack is set, so just run the callback. */
			rpnt->callback(ftp, 0, R_SUCCESS, NULL, NULL, NULL, FTP_OBJECT(rpnt));
		}

		/* state is set to SUCCESS and the number of requests awaiting to be sent out is decreased. */
		rpnt->state = R_SUCCESS;
		ftp->n_requests--;
	}
	
	return 1;
}


ftp_request *
ftp_request_add(_ftp_item *ftp, int rpos, ftp_request *rmain, int condition, FTP_CALLBACK(callback), const char *cmd_format, ...)
{
	char buf[GOOD_SIZE + 1];
	char *cmd = NULL, *args = NULL;
	va_list val;
	ftp_request **rpnt = NULL, *next = NULL, *prior = NULL, *tree_prior = NULL, *tree_next = NULL;
	

	memset(buf, 0, sizeof(buf));
	if (cmd_format) {
		va_start(val, cmd_format);
		vsnprintf(buf, GOOD_SIZE, cmd_format, val);
		va_end(val);
	}

	/* split into command and args. */
	cmd = buf;
	if ( (args = strchr(buf, ' ')) != NULL) {
		*args = 0;
		args++;
	}
	
	/* go to the correct position */
	rpnt = &ftp->requests;

	if (rmain != NULL && rpos == RP_END) {
		rpnt = &rmain;
		while (*rpnt) {
			tree_prior = *rpnt;
			rpnt = &(*rpnt)->tree_next;
		}
	} else if (rmain != NULL && rpos == RP_TREE_NEXT) {
		tree_prior = rmain;
		rpnt = &(rmain->tree_next);
		tree_next = rmain->tree_next;
		if (rmain->tree_next) {
			rmain->tree_next->tree_prior = *rpnt;
		}
	} else if (rmain == NULL && rpos == RP_END) {
		while (*rpnt) {
			prior = *rpnt;
			rpnt = &(*rpnt)->next;
		}
	} else if (rmain != NULL && rpos == RP_NEXT) {
		rpnt = &rmain;
		while ((*rpnt)->tree_prior)
			rpnt = &(*rpnt)->tree_prior;
		
		prior = *rpnt;
		next = (*rpnt)->next;
		rpnt = &(*rpnt)->next;
		
		if (*rpnt) {
			(*rpnt)->prior = prior; 
		}
	}

	/* allocate the entry. */
	if ( (*rpnt = (ftp_request *) j_malloc(sizeof(ftp_request))) == NULL) {
		MEM_ERR("malloc", ERR_SHOW);
		return NULL;
	}
	
	memset(*rpnt, 0, sizeof(ftp_request));
	
	(*rpnt)->cmd = NULL;
	if (cmd != NULL)
		(*rpnt)->cmd = j_strdup(cmd);

	(*rpnt)->args = NULL;
	if (args != NULL)
		(*rpnt)->args = j_strdup(args);

	(*rpnt)->type = REQ_SEND;
	(*rpnt)->condition = condition == 0 ? R_SUCCESS : condition;
	(*rpnt)->state = R_QUEUED;
	(*rpnt)->callback = callback;
	(*rpnt)->objects = NULL;
	(*rpnt)->next = next;
	(*rpnt)->tree_next = tree_next;
	(*rpnt)->tree_prior = tree_prior;
	(*rpnt)->prior = prior;

	ftp->n_requests++;
	gui_ftp_setup_write(ftp, 1);

	return *rpnt;
}


int
ftp_request_move(_ftp_item *ftp, ftp_request *from, ftp_request *to)
{
	ftp_request **fprior = NULL, **fnext = NULL;
	ftp_request **tprior = NULL, **tnext = NULL;
	ftp_request *last = NULL;
	/*
	 * ftype: "from" type
	 * ttype: "to" type
	 * values: 0: root request, 1: subrequest
	 */
	int ftype = 0, ttype = 0;
	

	if (!ftp || !from || !to)
		return 0;

	/* setup "from" entry */
	if (from->tree_prior == NULL) {
		fprior = &from->prior;
		fnext = &from->next;
		ftype = 0;
	} else {
		fprior = &from->tree_prior;
		fnext = &from->tree_next;
		ftype = 1;
	}

	/* setup "to" entry */
	if (to->tree_prior == NULL) {
		tprior = &to->prior;
		tnext = &to->next;
		ttype = 0;
	} else {
		tprior = &to->tree_prior;
		tnext = &to->tree_next;
		ttype = 1;
	}

	/* unlink from the chain */
	if (!ftype) {
		/* unlink the from entry from the original point */
		if (*fprior)
			(*fprior)->next = *fnext;
			
		if (*fnext)
			(*fnext)->prior = *fprior;
	} else {
		/* subrequest */
		if (*fprior)
			(*fprior)->tree_next = NULL;
	}

	if (!ttype) {
		/*
		 * "to type" is 0 (a root request)
		 */
		if (*tprior)
			(*tprior)->next = from;
			
		if (*tnext)			
			(*tnext)->prior = from;
		from->prior = *tprior;
		from->next = to;
	} else {
		last = from;
		while (last)
			last = last->tree_next;

		if (*tprior)
			(*tprior)->tree_next = from;

		last->tree_next = to;
		to->tree_prior = last;
		
	}
		
	return 1;
}

int
ftp_request_set_type(ftp_request *req, int type)
{
	if (!req)
		return -1;
	
	req->type = type;

	return 1;
}
