/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: ftp_resolver.c,v 1.10 2003/03/03 02:16:33 tgucio Exp $
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
/* local headers */
#include <ftp.h>
#include <ftp_resolver.h>
#include <jfork.h>
#include <jlist.h>
#include <log.h>
#include <misc.h>
#include <mem.h>
#include <error.h>


void
ftp_resolver_jlist_free_cb(jslist *j)
{
	ftp_resolver_info *rinfo = NULL;
	
	if (!j || !j->data)
		return;

	rinfo = j->data;
	ftp_resolver_info_free(rinfo);
	return;
}

/*
 * resolver callback, is called when a child has finished getting information from gethostbyname/getaddrinfo.
 * wraps up information sent from the child and calls a callback set for the ftp_resolver_lookup function.
 */
void
ftp_resolver_lookup_cb(jfork *item, void *notused)
{
	cb callback;
	void *data = NULL;
	int result;
	ftp_resolver_info *rinfo = NULL;
	jslist *jlist = NULL;
	
	DEBUG_MSG(DBG_INFO, "ftp_resolveR_lookup_cb");
	if (!item)
		return;

	// get the callback info
	callback = ftp_object_get_data(FTP_OBJECT(item), "ftp_resolver_callback");
	ftp_object_unlink_tag(FTP_OBJECT(item), "ftp_resolver_callback");

	// get the data info
	data = ftp_object_get_data(FTP_OBJECT(item), "ftp_resolver_data");
	ftp_object_unlink_tag(FTP_OBJECT(item), "ftp_resolver_data");
	
	jfork_read(item, (void *)&result, sizeof(int));
	// no ftp or no callback, no purpose to continue. (or lookup failed).
	if (!callback || result) {
		DEBUG_MSG(DBG_ERROR, "!callback || result not gotten");
		if (callback)
			callback(NULL, data);

		return;
	}
	
	result = 0;
	// build up the lookup return table
	while (1) {
		rinfo = j_malloc(sizeof(ftp_resolver_info));
		
		// read the length of the addr
		if (jfork_read(item, (void *)&rinfo->addr_len, sizeof(size_t)) <= 0)
			break; 

		if (rinfo->addr_len <= 0) 
			break; 

		// allocate space for the addr
		rinfo->addr = j_malloc(rinfo->addr_len);

		// read addr
		if (jfork_read(item, (void *)rinfo->addr, rinfo->addr_len) <= 0) {
			result = 1;
			break; // error
		}

		// read ai family
		if (jfork_read(item, (void *)&rinfo->ai_family, sizeof(int)) <= 0) {
			result = 1;
			break;	// error
		}
		
		// read protocol
		if (jfork_read(item, (void *)&rinfo->ai_proto, sizeof(int)) <= 0) {
			result = 1;
			break;
		}

		
		jlist = jslist_append(jlist, rinfo);
	}

	DEBUG_MSG(DBG_INFO, "mmmm, result is %d", result);
	callback(!result ? jlist : NULL, data);	
	jfork_close(item);

	/* free stuff */
	jslist_free_list(jlist, ftp_resolver_jlist_free_cb);

	return;
}


/*
 * set the port for info
 */
void
ftp_resolver_info_set_port(ftp_resolver_info *info, unsigned short port)
{
	if (!info || !info->addr)
		return;
	
	/* set the port */
	((struct sockaddr_in *)info->addr)->sin_port = htons(port);
	return;
}

/*
 * free allocated data in a resolver info item.
 */
void
ftp_resolver_info_free(ftp_resolver_info *rinfo)
{
	if (!rinfo)
		return;
		
	j_free(rinfo->addr);
	j_free(rinfo);
	return;
}




/*
 * resolve a host, non-blocking, and call the callback when finished.
 */
int
ftp_resolver_lookup(char *host, cb callback, void *data)
{
	jfork *pinfo = NULL;
	int result;
#ifdef IPV6 /* for all the cool people ;P */
	struct addrinfo hints, *ailist = NULL, *ainode = NULL;
#else
	struct hostent *he = NULL;
#endif
/*
 * XXX make it possible to use without forking?
 */

	/* setup a jfork. */
	if (!(pinfo = jfork_create(ftp_resolver_lookup_cb, NULL)))
		return 1;

	if (jfork_get_type(pinfo) == JFORK_PARENT) {
		ftp_object_set_data(FTP_OBJECT(pinfo), "ftp_resolver_callback", callback);
		ftp_object_set_data(FTP_OBJECT(pinfo), "ftp_resolver_data", data);
		return 0;
	}
	

#ifdef IPV6
	// setup hints. 
	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_flags=AI_PASSIVE;
	hints.ai_family=PF_UNSPEC;
	hints.ai_socktype=SOCK_STREAM;

	// run getaddrinfo, result 0 is success, 1 is failure
	result = 0;
	if (getaddrinfo(host, NULL, &hints, &ailist) != 0) {
		SYS_ERR("getaddrinfo", ERR_SHOW);
		result = 1;
	}

	// write out the result
	jfork_write(pinfo, &result, sizeof(int));

	// send the ailist info
	ainode = ailist;
	while (ainode) {
		// write length of addr 
		jfork_write(pinfo, (void *)&ainode->ai_addrlen, sizeof(size_t));
		
		// write addr
		jfork_write(pinfo, (void *)ainode->ai_addr, ainode->ai_addrlen);

		// write ai_family
		jfork_write(pinfo, (void *)&ainode->ai_family, sizeof(int));

		// write protocol
		jfork_write(pinfo, (void *)&ainode->ai_protocol, sizeof(int));
	
		ainode = ainode->ai_next;
	}
	
	freeaddrinfo(ailist);

#else 
	// look up host, result 0 is success, 1 is failure
	result = 0;
	if ( (he = gethostbyname(host)) == NULL) {
		SYS_ERR("gethostbyname", ERR_SHOW);
		result = 1;
	}
	
	// write out the result of the lookup
	jfork_write(pinfo, (void *)&result, sizeof(int));

	// if successful, send the addresses found
	if (!result) {
		char **hnode = NULL;
		struct sockaddr_in sin;
		size_t sin_size = sizeof(struct sockaddr_in);
		int ai_family = AF_INET;
		int ai_proto = IPPROTO_TCP;

		memset(&sin, 0, sizeof(struct sockaddr_in));
		// sin.sin_port = htons(ftp->port);
		sin.sin_family = PF_INET;
	
		// build the sin info and write to pipe
		hnode = he->h_addr_list;
		while (hnode && *hnode) {
			// write addr length
			jfork_write(pinfo, (void *)&sin_size, sizeof(size_t));

			// write addr
			memcpy(&sin.sin_addr, *hnode, he->h_length);
			jfork_write(pinfo, (void *)&sin, sin_size);

			// write ai_family
			jfork_write(pinfo, (void *)&ai_family, sizeof(int));

			// write protocol
			jfork_write(pinfo, (void *)&ai_proto, sizeof(int));
			
			hnode++;
		}
	}
		
#endif /* end of #ifdef IPV6 -> #else */
		
	jfork_close(pinfo);

	return 0;	// wont be reached (child exits).
}

