/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: ftp_transfer.c,v 1.7 2003/02/24 04:10:39 thepin Exp $
 */


#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <math.h>
/* local headers */
#include <config.h>
#include <defaults.h>
#include <error.h>
#include <ftp.h>
#include <ftp_object.h>
#include <ftp_transfer.h>
#include <jtimer.h>
#include <log.h>
#include <mem.h>
#include <misc.h>


/*
 * what we want to calculate:
 * ---------------------------
 *
 *    a) speed
 *       ~~~~~
 *	 find:
 *		speed of transfer (v_transfer (bytes/second))
 *
 *       variables: 
 *		(delta) change in bytes transferred (ds_transferred (bytes)),
 *		(delta) change in time it took to transfer s_transferred (dt_elapsed (seconds))
 *
 *       "algo:
 *		v_transfer = ds_transferred / dt_elapsed
 *
 *
 *    b) eta (estimated time of arrival)
 *       ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *	 find:
 *		estimated time of arrival (t_eta (seconds))
 *
 *       variables:
 *		total size of file (s_file (bytes)),
 *		speed of transfer (v_transfer (bytes / second))
 *
 *	"algo":
 *		t_eta = s_file / v_transfer
 *
 *
 *    c) transfer limiting sleep
 *       ~~~~~~~~~~~~~~~~~~~~~~~~
 *	 find:
 *		time to sleep (t_sleep (seconds))
 *
 * what we want to store:
 * ----------------------
 *
 *	a) current speed (v)
 *	   ~~~~~~~~~~~~~~~~~~~
 *		the current speed is measured as t -> 0, however, our time interval is probably not going to be that low.
 *		we will calculate the current speed over undefined time periods, at certain "events," such as after a read,
 *		or something like that.
 *
 *	b) average speed (v_av)
 *         ~~~~~~~~~~~~~~~~~~~~~
 *		the average speed is simply the whole of the bytes transferred over the whole time it took to transfer them.
 *
 *	c) eta (t_eta_v), using current speed (v)
 *         ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *		the estimated time of arrival in seconds, calculated using v (current speed).
 *
 *	d) eta (t_eta_v_av), using average speed (v_av)
 *         ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *		the eta in seconds, calculated using v_av (average speed).
 *
 *	e),f) queue etas (with v and v_av)
 *            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *              queue estimated time of arrival is the time it takes for the whole queue, 
 *              the sum of the etas of all files in the queue
 *
 *	?) transfer limiting sleep (t_sleep)
 *         ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *		the time to sleep (to compensate for going over speed limit) in seconds, calculated using v (current speed).
 *
 *		setsockopt(fd, proto, SO_RCVBUF, &oldsize, sizeof(int))
 *
 *	+ Concering the transfer limiting , should we use set "buffer sizes" with respect to the maximum speed set?
 *	  how could the buffer sizes be calculated, and would it reflect in transfer speed closer to the speed limit?
 */


/*
 * create a new ftp_transfer object
 */
ftp_transfer *
ftp_transfer_new()
{
	ftp_transfer *transfer = NULL;

	transfer = j_malloc(sizeof(ftp_transfer));
	memset(transfer, 0, sizeof(ftp_transfer));
	transfer->state = FTP_TRANSFER_NONE;
	return transfer;
}

/*
 * free a previously allocated ftp_transfer object.
 */
ftp_transfer *
ftp_transfer_free(ftp_transfer *transfer)
{
	double *v_current = NULL;
	
        if ((v_current = ftp_object_get_data(FTP_OBJECT(transfer), "ftp_transfer_current_speed"))) {
		j_free(v_current);
		ftp_object_unlink_tag(FTP_OBJECT(transfer), "ftp_transfer_current_speed");
	}
			
	j_free(transfer);
	return NULL;
}



/*
 * reset the content of the ftp_transfer object
 */
void
ftp_transfer_reset(ftp_transfer *transfer)
{
	int ignore_config;
	double speed_limit;

	if (!transfer)
		return;
		
	// ignore config for transfer speeds?
	ignore_config = (int)ftp_object_get_data(FTP_OBJECT(transfer), "ignore_config");
	speed_limit = transfer->v_max;
	
	// free stuff
	ftp_object_free(FTP_OBJECT(transfer));
	memset(transfer, 0, sizeof(ftp_transfer));
	/* XXX free the timers? */

	// set again (some stuff should be used again)
	transfer->state = FTP_TRANSFER_NONE;
	transfer->v_max = speed_limit;
	ftp_object_set_data(FTP_OBJECT(transfer), "ignore_config", (void *)ignore_config);
	
	return;
}

/*
 * set transfer state
 */
void
ftp_transfer_set_state(ftp_transfer *transfer, int state)
{
	if (!transfer)
		return;
	
	transfer->state = state;
	return;
}

/*
 * get transfer state
 */
int
ftp_transfer_get_state(ftp_transfer *transfer)
{
	if (!transfer)
		return -1;
	
	return transfer->state;
}



/*
 * set the resume position of a transfer
 */
void
ftp_transfer_set_resume_position(ftp_transfer *transfer, size_t resume_position)
{
	if (!transfer || ftp_transfer_get_state(transfer) == FTP_TRANSFER_NONE)
		return;
	
	transfer->s_resume_pos = resume_position;
	return;
}


/*
 * get the resume position of a transfer
 */
size_t
ftp_transfer_get_resume_position(ftp_transfer *transfer)
{
	if (!transfer || ftp_transfer_get_state(transfer) == FTP_TRANSFER_NONE)
		return 0;
	
	return transfer->s_resume_pos;
}




/*
 * set the transfer position of a transfer
 */
void
ftp_transfer_set_position(ftp_transfer *transfer, size_t position)
{
	if (!transfer || position < 0)
		return;
	
	transfer->s_transferred = position;
	return;
}

/*
 * get the transfer position
 */
size_t
ftp_transfer_get_position(ftp_transfer *transfer)
{
	if (!transfer || ftp_transfer_get_state(transfer) == FTP_TRANSFER_NONE)
		return 0;
	
	return transfer->s_transferred;
}

/*
 * set transfer size
 */
void
ftp_transfer_set_size(ftp_transfer *transfer, size_t size)
{
	if (!transfer || size < 0)
		return;

	transfer->s_size = size;
	return;
}

/*
 * get transfer size
 */
size_t
ftp_transfer_get_size(ftp_transfer *transfer)
{
	if (!transfer || ftp_transfer_get_state(transfer) == FTP_TRANSFER_NONE)
		return 0;
	
	return transfer->s_size;
}

/*
 * prepare the data structure for the transmission,
 * create objects and initialize timers
 */
void
ftp_transfer_setup(ftp_transfer *transfer, int state)
{
	if (!transfer)
		return;
	
	// set the state (upload / download)
	ftp_transfer_set_state(transfer, state);

	// reset the number of transferred bytes
	ftp_transfer_set_position(transfer, 0);

	// if a timer does not exist, create it
	if (!transfer->timer)
		transfer->timer = jtimer_new();
	
	// start the timer
	jtimer_start(transfer->timer);

	return;
}

/*
 * get the time (in seconds) needed to sleep, to maintain the download limit
 */
double
ftp_transfer_get_sleep_time(ftp_transfer *transfer)
{
	double t_sleep, t_max;
	double v_max, v_current;
	unsigned int cmax;
	
	if (!transfer || ftp_transfer_get_state(transfer) == FTP_TRANSFER_NONE)
		return 0;

	/*
	 * ignore_config is set so that it will ignore the speed limiting set in the config
	 */
	 
	t_max = 5.0;
	v_max = transfer->v_max;
 	if (!ftp_object_get_data(FTP_OBJECT(transfer), "ignore_config")) {
		cmax = config_read_num("transfer-limit");
		if (v_max != (int)cmax) {
			/* transfer limit bound changed, reset speed infos */
			jtimer_start(transfer->timer);
			ftp_transfer_set_resume_position(transfer, ftp_transfer_get_position(transfer));
			transfer->v_max = (double)cmax;
				
			return 0.0;
		}
		
		if (v_max <= 0 && cmax > 0)
			v_max = (double)cmax;
	}
	
	if (v_max <= 0 || t_max <= 0)
		return 0.0;
	
	v_current = ftp_transfer_get_average_speed(transfer);
	if (v_current < v_max || isinf(v_current) || isinf(v_max))
		t_sleep = 0.0;
	else
		t_sleep =  (double)v_current / (double)v_max;
		  
	if (t_sleep > t_max)
		t_sleep = t_max;

	return (t_sleep > 0 ? t_sleep : 0);
}


/*
 * end of transfer, reset the data structure
 */
void
ftp_transfer_end(ftp_transfer *transfer)
{
	double *gamma = NULL;
	if (!transfer)
		return;

	if ((gamma = ftp_object_get_data(FTP_OBJECT(transfer->timer), "start_pos"))) {
		j_free(gamma);
		ftp_object_unlink_tag(FTP_OBJECT(transfer->timer), "start_pos");
	}
	
	ftp_transfer_reset(transfer);
	return;
}


/*
 * get the average speed of transfer
 * v_transfer = s_transferred / t_elapsed;
 */
double
ftp_transfer_get_average_speed(ftp_transfer *transfer)
{
	unsigned int s_transferred;
	double v_transfer, t_elapsed;
	double delta;
	
	if (!transfer || ftp_transfer_get_state(transfer) == FTP_TRANSFER_NONE)
		return 0.0;

	
	t_elapsed = jtimer_get_elapsed_time(transfer->timer);
	s_transferred = transfer->s_transferred;

	if (t_elapsed <= 0 || isinf(t_elapsed))
		return 0.0;
	
	delta = ftp_transfer_get_resume_position(transfer);
	v_transfer = (double)((double)s_transferred - (double)delta) / (double)t_elapsed;
	return v_transfer;
}


/*
 * get the current ("instantaneous") speed of transfer.
 */
double
ftp_transfer_get_current_speed(ftp_transfer *transfer)
{
	double *v_current = NULL;
	
	if (!transfer || ftp_transfer_get_state(transfer) == FTP_TRANSFER_NONE)
		return 0.0;

	if (!(v_current = ftp_object_get_data(FTP_OBJECT(transfer), "ftp_transfer_current_speed"))) {
		return ftp_transfer_get_average_speed(transfer);
	}

	if (*v_current < 0 || isinf(*v_current))
		return 0.0;

	return *v_current;
}


/*
 * calculate current speed
 * v_transfer = ds_transferred / dt_elapsed
 */
double
ftp_transfer_calculate_current_speed(ftp_transfer *transfer)
{
	jtimer *timer = NULL;
	unsigned int ds_transferred;
	double v_transfer, dt_elapsed;

	if (!transfer || ftp_transfer_get_state(transfer) == FTP_TRANSFER_NONE)
		return 0.0;
		
	timer = ftp_object_get_data(FTP_OBJECT(transfer), "ftp_transfer_delta_timer");
	ds_transferred = 
		ftp_transfer_get_position(transfer)
		- (unsigned int)ftp_object_get_data(FTP_OBJECT(transfer), "ftp_transfer_delta_bytecount");
		
	if (!timer || ds_transferred <= 0)
		return 0.0;
	
	dt_elapsed = jtimer_get_elapsed_time(timer);
	if (dt_elapsed <= 0 || isinf(dt_elapsed))
		return 0.0;
		
	v_transfer = (double) ds_transferred / (double) dt_elapsed;
	return v_transfer;
}




/*
 * update the transfer information
 */
void
ftp_transfer_update(ftp_transfer *transfer, size_t ds_bytes)
{
	jtimer *timer = NULL;
	double *current_speed = NULL;

	if (!transfer || ftp_transfer_get_state(transfer) == FTP_TRANSFER_NONE)
		return;


// <kill me>
//printf("current speed is: %0.4f kb/s\n", (double)ftp_transfer_get_current_speed(transfer) / (double)1024);
//printf("average speed is: %0.4f kb/s\n", (double)ftp_transfer_get_average_speed(transfer) / (double)1024);
//printf("sleep time is %0.4fs\n", ftp_transfer_get_sleep_time(transfer));
//printf("transfer size is %d\n", ftp_transfer_get_size(transfer));
// </kill me>

	/* update the figure of total bytes transferred */
	transfer->s_transferred += ds_bytes;

	/* some calculations regarding current speed. */
	timer = ftp_object_get_data(FTP_OBJECT(transfer), "ftp_transfer_delta_timer");
	if (!timer) {
		timer = jtimer_new();
		jtimer_start(timer);
	}
		
	if (jtimer_get_elapsed_time(timer) > 2.0) {
		current_speed = ftp_object_get_data(FTP_OBJECT(transfer), "ftp_transfer_current_speed");
		if (!current_speed) {
			current_speed = j_malloc(sizeof(double));
			ftp_object_set_data(FTP_OBJECT(transfer), "ftp_transfer_current_speed", current_speed);
		}
		
		*current_speed = ftp_transfer_calculate_current_speed(transfer);
		jtimer_start(timer);
		ftp_object_set_data(FTP_OBJECT(transfer), "ftp_transfer_delta_bytecount", (void *)transfer->s_transferred);
	}

	ftp_object_set_data(FTP_OBJECT(transfer), "ftp_transfer_delta_timer", timer);

	return;
}

