/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: gui_toolbar.c,v 1.3 2003/03/03 03:05:33 thepin Exp $
 */


#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <strings.h>
#include <time.h>
#include <glib.h>
#include <gtk/gtk.h>
/* local headers */
#include <defaults.h>
#include <gui.h>
#include <gui_toolbar.h>
#include <mem.h>



/*
 * called upon destruction
 */
void
gui_toolbar_destroy_cb(GObject *obj, gpointer data)
{
	gui_toolbar *toolbar = NULL;

	/* XXX Does this ever get called? */
//	DEBUG_MSG("gui_toolbar_destroy_cb, destroying the info for the toolbar\n");
	if (!(toolbar = data))
		return;

	gui_toolbar_free(toolbar);
	return;
}



/*
 * new toolbar object
 */
GtkWidget *
gui_toolbar_new()
{
	gui_toolbar *toolbar = NULL;

	/* allocate the menu bar */
	toolbar = j_malloc(sizeof(gui_toolbar));
	memset(toolbar, 0, sizeof(gui_toolbar));
	
	/* setup frame */
	toolbar->frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(toolbar->frame), 
			GTK_SHADOW_OUT);
	g_object_set_data(G_OBJECT(toolbar->frame), 
				"gui_toolbar", toolbar);
	
	/* setup hbox */
	toolbar->hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(toolbar->hbox);

	/* put hbox into frame */
	gtk_container_add(GTK_CONTAINER(toolbar->frame), toolbar->hbox);
	
	g_signal_connect(G_OBJECT(toolbar->frame), "destroy", 
		G_CALLBACK(gui_toolbar_destroy_cb), toolbar);


	return toolbar->frame;
}


/*
 * free toolbar object
 */
gui_toolbar *
gui_toolbar_free(gui_toolbar *toolbar)
{	
	GList *children = NULL;
	
	if (!toolbar)
		return NULL;

	children = toolbar->children;
	while (children) {
		gui_toolbar_child_free(children->data);
		children = g_list_next(children);
	}

	g_list_free(toolbar->children);
	j_free(toolbar);
	return NULL;
}


/*
 * new toolbar child object
 */
gui_toolbar_child *
gui_toolbar_child_new()
{
	gui_toolbar_child *child = NULL;

	/* allocate and null toolbar child*/
	child = j_malloc(sizeof(gui_toolbar_child));
	memset(child, 0, sizeof(gui_toolbar_child));

	return child;
}

/*
 * free toolbar child object
 */
gui_toolbar_child *
gui_toolbar_child_free(gui_toolbar_child *child)
{	
	if (child)
		j_free(child);

	return NULL;
}



/* 
 * get the max requisition of all widgets in the toolbar
 */
static void
gui_toolbar_item_get_max_requisition(gui_toolbar *toolbar, GtkRequisition *requisition)
{
	GList *children = NULL;
	GtkRequisition child_req;
	gui_toolbar_child *child = NULL;
	gint width = 0, height = 0;
	
	if (!toolbar || !requisition)
		return;

	children = toolbar->children;
	while (children) {
		child = children->data;
		if (child && child->resize == TRUE) {
			gtk_widget_size_request(child->widget, &child_req);
			if (child_req.height > height)
				height = child_req.height;
			
			if (child_req.width > width)
				width = child_req.width;
		}
		
		children = g_list_next(children);
	}
	
	requisition->width = width;
	requisition->height = height;
	
	return;
}


/*
 * set all the resizable childrens to the size of the greatest
 * resizable tool in the toolbar
 */
static void
gui_toolbar_resize(GtkWidget *mainframe)
{
	GtkRequisition requisition;
	GList *children = NULL;
	gui_toolbar *toolbar = NULL;
	gui_toolbar_child *child = NULL;

	toolbar = g_object_get_data(G_OBJECT(mainframe), "gui_toolbar");
	if (!toolbar || !toolbar->hbox)
		return;

	gui_toolbar_item_get_max_requisition(toolbar, &requisition);
	if (requisition.height <= 0 || requisition.width <= 0)
		return;
	
	children = toolbar->children;
	while (children) {
		child = children->data;
		if (child && child->resize == TRUE)
			gtk_widget_set_size_request(child->widget, requisition.width, requisition.height);

		children = g_list_next(children);
	}

	return;
}



/*
 * append a widget to the toolbar
 */
void
gui_toolbar_append_widget(GtkWidget *mainframe, int alignment, int resize, GtkWidget *widget)
{
	gui_toolbar *toolbar = NULL;
	gui_toolbar_child *child = NULL;
	
	toolbar = g_object_get_data(G_OBJECT(mainframe), "gui_toolbar");
	if (!toolbar || !toolbar->hbox)
		return;

	/* create a new child */
	child = gui_toolbar_child_new();
	child->widget = widget;
	child->alignment = alignment;
	child->resize = resize;
	
	/* pack into the main box */
	if (alignment == GUI_JUSTIFY_LEFT) {
		gtk_box_pack_start(GTK_BOX(toolbar->hbox), child->widget, FALSE, FALSE, 0);
	} else {
		gtk_box_pack_end(GTK_BOX(toolbar->hbox), child->widget, FALSE, FALSE, 0);
	}
	gtk_widget_show(toolbar->hbox);

	/* throw the child into the list of children*/
	toolbar->children = g_list_insert(toolbar->children, child, toolbar->n_children++);

	gtk_widget_show_all(child->widget);
	gui_toolbar_resize(mainframe);
	
	return;
}



/*
 * append to the toolbar
 */
void
gui_toolbar_append_item(GtkWidget *mainframe, int alignment, const char *text, GtkWidget *icon, GtkSignalFunc callback, gpointer user_data)
{
	GtkWidget *button = NULL;
	GtkWidget *vbox = NULL;
	GtkWidget *label = NULL;
	
	button = gtk_button_new();
	gtk_button_set_relief (GTK_BUTTON(button), GTK_RELIEF_NONE);
	GTK_WIDGET_UNSET_FLAGS (button, GTK_CAN_FOCUS);
	vbox = gtk_vbox_new(FALSE, 0);
	
	/* set label */
	if (text) {
		label = gtk_label_new(text);
		gtk_box_pack_end(GTK_BOX(vbox), label, FALSE, FALSE, 0);
		gtk_widget_show(label);
	}
	
	/* set an icon */
	if (icon) {
		gtk_box_pack_end(GTK_BOX(vbox), icon, FALSE, FALSE, 0);
		gtk_widget_show(icon);
	}
	
	/* setup the callback for the button */
	if (callback) {
		gtk_signal_connect(GTK_OBJECT(button), "clicked", callback, user_data);
	}
	
	/* put the widget onto the button */
	gtk_container_add(GTK_CONTAINER(button), vbox);

	/* put it on the main box */
	gui_toolbar_append_widget(mainframe, alignment, TRUE, button);

	return;
}
