/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: hammering.c,v 1.17 2003/03/07 21:22:12 thepin Exp $
 */


#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/stat.h>
/* local headers */
#include <ftp_event.h>
#include <ftp_request.h>
#include <ftp.h>
#include <ftp_misc.h>
#include <ftp_cmd.h>
#include <ftp_net.h>
#include <config.h>
#include <cache.h>
#include <medium.h>
#include <error.h>
#include <log.h>
#include <mem.h>
#include <hammering.h>

static void (*signal_hammering_start)(_ftp_item *) = NULL;
static void (*signal_hammering_stop)(_ftp_item *) = NULL;
static void (*signal_hammering_connected)(_ftp_item *) = NULL;

#define SIGNAL_HAMMERING(name, arg) 		\
	if (signal_hammering_ ## name)	 	\
		signal_hammering_ ## name (arg)

/*
 * remove the gui stuff.. maybe add some functions that can be set..
 * and are called if they are set.
 */

void
hammering_stop(_ftp_item *ftp)
{
	int ret;

	if (ftp_object_get_data(FTP_OBJECT(ftp), "hammering_enabled")) {
		medium_cmd(M_TIMEOUT_REMOVE, &ret, ftp_object_get_data(FTP_OBJECT(ftp), "hammering_timeout"));
	}
	
	FTP_MISC_LOG(ftp, "    Stopped hammering\n");
	
	ftp_object_unlink_tag(FTP_OBJECT(ftp), "hammering_enabled");
	ftp_object_unlink_tag(FTP_OBJECT(ftp), "hammering_timeout");

	SIGNAL_HAMMERING(stop, ftp);

	return;
}

int
hammering_enabled(_ftp_item *ftp)
{
	return (int)ftp_object_get_data(FTP_OBJECT(ftp), "hammering_enabled");
}





void
hammering_connected_cb(_ftp_item *ftp)
{
	ftp_request *req = NULL;
	
	DEBUG_MSG(DBG_INFO, "hammering_connected_cb");
	if (!ftp) 
		return;
	
	if (!(req = ftp_login(ftp)))
		return;
		
	ftp_pwd_with_req(ftp, req);

	SIGNAL_HAMMERING(connected, ftp);
	ftp_object_unlink_tag(FTP_OBJECT(ftp), "hammering_enabled");
	return;
}

/*
	here the actual connecting is done.
	The hammering will be more useful after new connect functions have
	been added into junkie and that process abstracted better, like the interaction
	between the gui and the ftp lib for connecting. XXX FIXME
	When that is ready, the gui* stuff can be removed from here.
*/
void
hammering_timeout_callback(void *data)
{
	_ftp_item *ftp = NULL;
//	ftp_request *req = NULL;


	DEBUG_MSG(DBG_INFO, "hammering_timeout_callback");
	if (!(ftp = data))
		return;

	DEBUG_MSG(DBG_INFO, "calling ftp_reconnect");
	if (!ftp_reconnect(ftp, hammering_connected_cb, NULL)) {
		DEBUG_MSG(DBG_INFO, "failed to reconnect");
		ftp_free(&ftp);
		return;
	}

	
	return;
}

/*
** This function is a callback for the ftp USER/PASS commands.  If the either of those command was not issued succesfully,
** and the hammering toggled on, junkie will try to reconnect after certain time specified in the preferences.
*/
FTP_EVENT_CB_CMD(hammering_cb)
{
	int delay, ret;

	DEBUG_MSG(DBG_INFO, "hammering_cb");
	if (!ftp)
		return FTP_EVENT_FAILURE;
	
	if (ftp_object_get_data(FTP_OBJECT(ftp), "hammering_enabled"))
		return FTP_EVENT_SUCCESS;

	/*
	 * find a better method to see if we should hammer or not.
	 */
	if (code != 421/* && code != 530*/)
		return FTP_EVENT_SUCCESS;

	/* Stop if the hammering support is not enabled. */
	if ( (delay = config_read_num("hammering")) <= 0)
		return FTP_EVENT_SUCCESS;

	if (medium_cmd(M_TIMEOUT_ADD, &ret, delay*1000, hammering_timeout_callback, ftp)) {
		SIGNAL_HAMMERING(start, ftp);
		FTP_MISC_LOG(ftp, "    Hammering is enabled, going to retry after %d seconds\n", delay);
		ftp_object_set_data(FTP_OBJECT(ftp), "hammering_enabled", (void *)1);
		ftp_object_set_data(FTP_OBJECT(ftp), "hammering_timeout", (void *)ret);
		ftp_close(ftp);
	}
	
	return FTP_EVENT_SUCCESS;
}

void
hammering_enable_handlers(_ftp_item *ftp)
{
	ftp_event_add_handler(ftp, FTP_EVENT_CMD, "USER", hammering_cb);
	ftp_event_add_handler(ftp, FTP_EVENT_CMD, "PASS", hammering_cb);
	return;
}

int
hammering_signal_connect(int signal, void (*func))
{
	switch (signal) {
	case HAMMERING_START:
		signal_hammering_start = func;
		break;
		
	case HAMMERING_STOP:
		signal_hammering_stop = func;
		break;

	case HAMMERING_CONNECTED:
		signal_hammering_connected = func;
		break;
	}

	return 0;
}

