/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: jio.c,v 1.6 2003/03/07 17:48:13 thepin Exp $
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
/* local headers */
#include <jio.h>
#include <mem.h>
#include <defaults.h>
#include <misc.h>

/*
 * new jio object
 */
jio * 
jio_new()
{
	jio *fh;
	
	fh = j_malloc(sizeof(jio));
	memset(fh, 0, sizeof(jio));
	return fh;
}

/*
 * del a jio object
 */
void
jio_free(jio *fh)
{
	if (!fh)
		return;
		
	j_free(fh->file);
	j_free(fh->temp);
	j_free(fh);
	return;
}


/*
 * open a file and return an jio fd object
 * if O_TRUNC flag is set, then open a temporary file, which is then renamed on close.
 */
jio *
j_open(const char *pathname, int flags, mode_t mode)
{
	jio *fh = jio_new();
	
	fh->file = j_strdup(pathname);	
	fh->trunc = flags & O_TRUNC & (O_RDWR | O_WRONLY);
	
	if (fh->trunc) {
		fh->temp = j_asprintf("%s.tmp", pathname);
		if (mode)
			fh->fd = open(fh->temp, flags, mode);
		else
			fh->fd = open(fh->temp, flags);
	}
	else {
		if (mode)
			fh->fd = open(pathname, flags, mode);
		else
			fh->fd = open(pathname, flags);
	}
	
	if(fh->fd == -1) {
		jio_free(fh);
		perror("(j_open) failed to open file: ");
		return NULL;
	}

	return fh;
}


/*
 * read from a jio channel
 */
size_t
j_read(jio *fh, void *buf, size_t count)
{
	size_t r;

	if (!fh)
		return 0;
		
	if ((r = read(fh->fd, buf, count)) == -1) {
		fh->trunc = 0;
		perror("(j_read) failed to read from file: ");
	}
	
	return r;
}


/*
 * write to a jio channel
 */
size_t
j_write(jio *fh, void *buf, size_t count)
{
	size_t r;
	
	if (!fh)
		return 0;
		
	r = write(fh->fd, buf, count);
	
	if (r == -1) {
		fh->trunc = 0;
		perror("(j_write) failed to write to file: ");
	}
	
	return r;
}

/*
 * write formatted output to a jio channel
 */
size_t
j_printf(jio *fh, char *format, ...)
{
	va_list ap;
	char buf[MAX_SIZE];
	
	va_start(ap, format);	
	vsnprintf(buf, MAX_SIZE -1, format, ap);
	
	return j_write(fh, buf, strlen(buf));
}


/*
 * close a jio channel
 */
int
j_close(jio *fh)
{
	int r;
	
	if (!fh)
		return 0;
		
	r = close(fh->fd);
	
	if(fh->trunc)
		r = rename(fh->temp, fh->file);
		
	if(r == -1)
		perror("(j_close): ");

	jio_free(fh);	
	return r;
}
	
