/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: local.c,v 1.32 2003/03/07 17:48:13 thepin Exp $
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>

/* local headers */
#include <ftp.h>
#include <ftp_cmd.h>
#include <ftp_object.h>
#include <ftp_event_callback.h>
#include <ftp_request.h>
#include <misc.h>
#include <error.h>
#include <local.h>
#include <mem.h>

#ifndef PATH_MAX
#define PATH_MAX	4096		// XXX /usr/include/limits.h:#define _POSIX_PATH_MAX         255
#endif

/*
** Constructs a permission "string" from an permission value.
** pretty much the same as strmode(3).
*/
char *
get_perm(struct stat *stat)
{
	char *buf = NULL, *pnt = NULL;
	char c;

	if ( (pnt = buf = (char *)j_malloc(sizeof(char)*11)) == NULL) 
		return NULL;

	/*
		first character can be:
		b: block special file
		c: character special file
		d: directory
		l: symbolic link
		s: socket link
		p: fifo
		-: regular file
		owner, group, others: --- (3x3)
		owner, group:
		r: readable  -: not readable
		w: writable  -: not writable
		 x: executable -: not executable (S, s are possible too)
		 s: suid file is exec, (if in group, set-group-ID mode is set)
		 S: same as s, but not executable
		others:
		same for r,w
		for the 3rd, the same applies except that T,t is added
		T: STicky bit is set, but not execute or search permission
	        t: sticky bit set, execute and search permission	
	*/

	/* first char */
	if (S_ISBLK(stat->st_mode)) c = 'b';
	else if (S_ISCHR(stat->st_mode)) c = 'c';
	else if (S_ISDIR(stat->st_mode)) c = 'd';
	else if (S_ISLNK(stat->st_mode)) c = 'l';
	else if (S_ISSOCK(stat->st_mode)) c = 's';
	else if (S_ISFIFO(stat->st_mode)) c = 'p';
	else if (S_ISREG(stat->st_mode)) c = '-';
	else c = '?';

	*(pnt++) = c;
	
	/* owner ??? */
	*(pnt++) = (stat->st_mode & S_IRUSR) ? 'r' : '-';
	*(pnt++) = (stat->st_mode & S_IWUSR) ? 'w' : '-';
	if (stat->st_mode & S_IXUSR) {
		if (stat->st_mode & S_ISUID)
			*(pnt++) = 's';
		else
			*(pnt++) = 'x'; 
	} else {
		if (stat->st_mode & S_ISUID)
			*(pnt++) = 'S';
		else
			*(pnt++) = '-';
	}

	/* group */
	*(pnt++) = (stat->st_mode & S_IRGRP) ? 'r' : '-';
	*(pnt++) = (stat->st_mode & S_IWGRP) ? 'w' : '-';
	if (stat->st_mode & S_IXGRP) { 
		if (stat->st_mode & S_ISUID)
			*(pnt++) = 's';
		else
			*(pnt++) = 'x';
	} else {
		if (stat->st_mode & S_ISUID)
			*(pnt++) = 'S';
		else
			*(pnt++) = '-'; 
	}

	/* others */
	*(pnt++) = (stat->st_mode & S_IROTH) ? 'r' : '-';
	*(pnt++) = (stat->st_mode & S_IWOTH) ? 'w' : '-';
 	if (stat->st_mode & S_IXOTH) {
		if (stat->st_mode & S_ISVTX)
			c = 't';
		else
			c = 'x';
	} else {
		if (stat->st_mode & S_ISVTX)
			c = 'T';
		else
			c = '-';
	}
	*(pnt++) = c;
	*pnt = 0;
 
	return buf;
}


int
local_cwd(const char *path)
{
	int i = chdir(path);
	
	if(i)
		perror("chdir");
		
	return i;
}


char *
local_pwd()
{
	static char cwd[PATH_MAX +1];
	if(!getcwd(cwd, PATH_MAX))	return NULL;
	return cwd;
}

void
local_cdup()
{
	chdir("..");
	return;
}

struct _list *
local_list()
{
	DIR *dir = NULL;
	struct _list *list_dir = NULL;
	struct _list **list_pnt = NULL;
	struct dirent *ent = NULL;
	struct stat stats;
	struct passwd *owner = NULL;
	struct group *grp = NULL;
	char cwd[PATH_MAX + 1];

	if (! getcwd(cwd, PATH_MAX) ) {
		SYS_ERR("getcwd", ERR_SHOW);
		return NULL;
	}

	if (! (dir = opendir(cwd)) )
		return NULL;

	list_pnt = &list_dir;

	while( (ent = readdir(dir)) ) {
		if (!strcmp(ent->d_name, ".") || !strcmp(ent->d_name, ".."))
			continue;

		if ( (*list_pnt = (struct _list *)j_malloc(sizeof(struct _list))) == NULL) 
			return NULL;

		(*list_pnt)->next = 0;
		(*list_pnt)->child = 0;

		(*list_pnt)->file = j_strdup(ent->d_name);
		/* XXX strdup error check */
		
		stat(ent->d_name, &stats);

		(*list_pnt)->size = stats.st_size;
		(*list_pnt)->type = S_ISDIR(stats.st_mode) ? _LIST_DIR : _LIST_FILE;
		(*list_pnt)->perm = get_perm(&stats);
		(*list_pnt)->mtime = stats.st_mtime;

		/* owner */
		if ( (owner = getpwuid(stats.st_uid)) == NULL)
			(*list_pnt)->owner = NULL;
		else
			(*list_pnt)->owner = j_strdup(owner->pw_name);
			
		/* group */
		if ( (grp = getgrgid(stats.st_gid)) == 0)
			(*list_pnt)->group = 0;
		else
			(*list_pnt)->group = j_strdup(grp->gr_name);
		
		list_pnt = (struct _list **)&((*list_pnt)->next);
	}
	closedir(dir);

	return list_dir;
}

int local_mv(char *src, char *dest)
{
	int n;
	char* new;
	
	new = j_asprintf("%s/%s",dest,src);
	n = rename(src, new);
	if (n == -1) {
		DEBUG_MSG(DBG_INFO, "rename failed, %s", strerror(errno));
		SYS_ERR("rename", ERR_SHOW);
		return -1;
	}
	
	j_free(new);

	return 1;
}

void
local_remove(char *path)
{
	DIR *dir;
	struct dirent *ent;
	char *file;
	
	if (remove(path) == 0)
		return;
		
	if (errno == ENOTEMPTY) {
		if(!(dir = opendir(path))) {
			DEBUG_MSG(DBG_INFO, "opendir failed, %s", strerror(errno));
			return;
		}
		
		while( (ent = readdir(dir)) ) {
			if(!strcmp(ent->d_name, ".") || !strcmp(ent->d_name, ".."))
				continue;
			else {
				file = j_asprintf("%s/%s", path, ent->d_name);
				local_remove(file);
				j_free(file);
			}
		}
		
		closedir(dir);
	} else 
		DEBUG_MSG(DBG_INFO, "%s", strerror(errno));
	
	
	remove(path);
	j_free(path);
	
	return;
}

						

void
local_mkdir(const char *newdir)
{
	char *file;
	char cwd[MED_SIZE + 1];
	
	if (!getcwd(cwd, MED_SIZE))
		return;
		
	file = j_asprintf("%s/%s", cwd, newdir);
	mkdir(file, 0777);
	j_free(file);
	
	return;
}

int
local_mkpath(char *localfile)
{
	size_t len = 0;
	char *j = NULL;
	struct stat _stat;
	char *path = j_strdup(localfile);
	
	len = strlen(path);
	do {
		j = strrchr(path, '/');
		if(j)
			*j = '\0';
		else
			break;
	} while(stat(path, &_stat) == -1);
	
	j = strrchr(path, '/');
	if(j)   *j = '\0';
	
	while( (path[strlen(path)] = '/') ) {
		if(strlen(path) >= len)	break;

		mkdir(path, 0777);
	}
	
	j_free(path);
	return 1;
}
		

FTP_CB_CMD(local_free_list_items)
{
	struct _list *list_dir, *local_dir;
	list_dir  = ftp_object_get_data(obj, "list_dir");
	local_dir = ftp_object_get_data(obj, "local_dir");

	j_free(list_dir->file); j_free(list_dir);
	if (local_dir != list_dir)
		j_free(local_dir->file), j_free(local_dir);
		
	return;
}
					

void
localdir_stor(_ftp_item *ftp, struct _list *list_dir)
{
	struct _list *dup = j_malloc(sizeof(struct _list));
	dup->file = j_strdup(list_dir->file);

	localdir_stor_handler(ftp, dup, 0);
}
	

void
localdir_stor_handler(_ftp_item *ftp, 
		                struct _list *list_dir, 
	 	                struct _list *local_dir)
{
	ftp_request *req;

	if(!local_dir) local_dir = list_dir;
	
	req = ftp_mkdir_wrapper(ftp, RP_END, NULL, 0, localdir_stor_cb, list_dir->file);
	ftp_object_set_data(FTP_OBJECT(req), "local_dir", local_dir);
	ftp_object_set_data(FTP_OBJECT(req), "list_dir", list_dir);
	
}

FTP_CB_CMD(localdir_stor_cb)
{
	ftp_request *req;
	DIR * dir;
	struct dirent *ent;
	struct stat stats;
	struct _list *dp, *dq;
	struct _list *local_dir, *list_dir;
	local_dir = ftp_object_get_data(obj, "local_dir");
	list_dir  = ftp_object_get_data(obj, "list_dir");
	
	if(! (dir = opendir(local_dir->file)))	return ;
	
	
	while( (ent = readdir(dir)) ) {

		if(!strcmp(ent->d_name, ".") || !strcmp(ent->d_name, ".."))
			continue;

		dp = j_malloc(sizeof(struct _list));
		dq = j_malloc(sizeof(struct _list));
				
		// ?? asprintf(&dp->file, "%s/%s", list_dir->file,  ent->d_name);
		dp->file = j_asprintf("%s/%s", local_dir->file, ent->d_name);
		// where is this j_free()d? 

		stat(dq->file, &stats);
		dp->size = dq->size = stats.st_size;
		
		/* XXX error check and set some other fields ? */
		if(S_ISDIR(stats.st_mode)) {
			localdir_stor_handler(ftp, dp, dq);
			req = ftp_request_add(ftp, RP_NEXT, NULL, 0, local_free_list_items, 0);
			ftp_object_set_data(FTP_OBJECT(req), "local_dir", local_dir);
			ftp_object_set_data(FTP_OBJECT(req), "list_dir",  list_dir);
		}
		else {
			DEBUG_MSG(DBG_INFO, "added request for: %s\n", dp->file);
			req = ftp_stor(ftp, RP_END, NULL, 0, local_free_list_items, dp, dp->file, 0);
			ftp_object_set_data(FTP_OBJECT(req), "local_dir", dp);
			ftp_object_set_data(FTP_OBJECT(req), "list_dir",  dq);
				
		}
	}
	
	closedir(dir);
	return;
}



void
localdir_retr(_ftp_item *ftp, struct _list *list_dir)
{
	struct _list *dup = j_malloc(sizeof(struct _list));
	
	dup->file = j_strdup(list_dir->file);
	localdir_retr_handler(ftp, dup, 0);
	return;
}

void
localdir_retr_handler(_ftp_item *ftp,
		                struct _list *list_dir,
							 struct _list *local_dir)
{
	ftp_request *req;
		
	if(!local_dir)	local_dir = list_dir;
	
	if(mkdir(local_dir->file, 0777) )	return;
	
	req = ftp_list_with_cb(ftp, localdir_retr_cb, 0, list_dir->file); 
	ftp_object_set_data(FTP_OBJECT(req), "local_dir", local_dir);
	ftp_object_set_data(FTP_OBJECT(req), "list_dir",  list_dir);
}

FTP_CB_CMD(localdir_retr_cb)
{
	ftp_request *req;
	struct _list *list;
	struct _list *dp, *dq;
	struct _list *local_dir, *list_dir;
	local_dir = ftp_object_get_data(obj, "local_dir");
	list_dir  = ftp_object_get_data(obj, "list_dir");
			
	
	list = ftp->dirinfo->list_dir;
	while (list) {
		if(!strcmp(list->file, ".") || !strcmp(list->file, ".."))
			continue;
		
		dp = j_malloc(sizeof(struct _list));
		dq = j_malloc(sizeof(struct _list));
		
		// asprintf(&dp->file, "%s/%s", list_dir->file,  list->file);
		dp->file = j_asprintf("%s/%s", local_dir->file, list->file);
		// where is this j_free()d ?
	
		dp->size = dq->size = list->size;
		
		if(J_ISDIR(list->type)) {
			localdir_retr_handler(ftp, dp, dq);
			req = ftp_request_add(ftp, RP_TREE_NEXT, ftp->requests, 0, local_free_list_items, 0);
			ftp_object_set_data(FTP_OBJECT(req), "local_dir", local_dir);
			ftp_object_set_data(FTP_OBJECT(req), "list_dir",  list_dir);
		}
		else {
			req = ftp_retr(ftp, RP_NEXT, ftp->requests->next, 0, NULL, dp, dq->file, TRFR_DEFAULT);
			ftp_object_set_data(FTP_OBJECT(req), "local_dir", dp);
			ftp_object_set_data(FTP_OBJECT(req), "list_dir",  dq);
		}

		list = list->next;
	}
	
	return;
}
