/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: mem.c,v 1.14 2003/03/07 18:02:25 ya- Exp $
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/stat.h>
/* local headers */
#include <defaults.h>
#include <error.h>
#include <log.h>
#include <mem.h>

mem_tracker *mem_head;	/* The beginning of the mmemory linked list. (I am a banana :)) */


mem_tracker *
mem_tracker_new()
{
	mem_tracker *ptr = NULL;

	if (!(ptr = malloc(sizeof(mem_tracker)))) {
		MEM_ERR("malloc", ERR_SHOW);
		return NULL;
	}
	
	memset(ptr, 0, sizeof(mem_tracker));
	return ptr;
}

/*
 * find an entry by address
 */
mem_tracker *
mem_tracker_find_entry(void *target)
{
	mem_tracker *pnt = NULL;

	pnt = mem_head;
	while (pnt) {
		if (pnt->ptr == target)
			return pnt;
			
		pnt = pnt->next;
	}

	return NULL;
}

int
mem_tracker_add_entry(void *ptr, size_t size)
{
	mem_tracker *ment = NULL;
	mem_tracker **pnt = NULL, *prior = NULL;
	
	if (!ptr)
		return 0;

	if (!(ment = mem_tracker_find_entry(ptr))) {
		ment = mem_tracker_new();
		if (!ment)
			return -1;
		
		/* Go to the back of the memory list. */
		pnt = &mem_head;
		while (*pnt) {
			prior = *pnt;
			pnt = &(*pnt)->next;
		}

		*pnt = ment;
		(*pnt)->next = NULL;
		(*pnt)->prior = prior;
	}
	
	/* Register the entry. */
	ment->ptr = ptr;
	ment->size = size;

	return 1;
}

/*
 * 0 if not found
 * 1 if removed
 */
int
mem_tracker_unlink_addr(void *ptr)
{
	mem_tracker *mtrack = NULL, *mprev = NULL, *mnext = NULL;

	if (!(mtrack = mem_tracker_find_entry(ptr)))
		return 0;
		
	/* unlink and free the entry */
	mprev = mtrack->prior;
	mnext = mtrack->next;
	if (mprev) 
		mprev->next = mnext;
	else
		mem_head = mem_head->next;
		
	if (mnext)
		mnext->prior = mprev;
		
	mtrack->prior = mtrack->next = NULL;
	mtrack->ptr = NULL;
	free(mtrack);
		
	return 1;
}


void
mem_tracker_set_addr(mem_tracker *mtrack, void *ptr)
{
	if (!mtrack)
		return;
	
	mtrack->ptr = ptr;
	return;
}

void
mem_tracker_set_size(mem_tracker *mtrack, size_t size)
{
	if (!mtrack)
		return;
	
	mtrack->size = size;
	return;
}


/*
** allocate some memory 
*/
void *
j_malloc(size_t size)
{
	void *ptr = NULL;

	/* Allocate the memory */
	if (!(ptr = malloc(size))) {
		MEM_ERR("malloc", ERR_SHOW);
		return NULL;
	}

	/* Add the entry to the memory tracker. */
	mem_tracker_add_entry(ptr, size);

	return ptr;
}

void *
j_realloc(void *ptr, size_t size)
{
	char *gamma = NULL;
	mem_tracker *ment = NULL;

/*
 * If size is 0, either a null pointer or a unique pointer
 * that can be successfully passed to free() is returned
 */
	gamma = realloc(ptr, size);
	if (!gamma)
		return NULL;
	
	ment = mem_tracker_find_entry(ptr);
	if (ment) {
		mem_tracker_set_size(ment, size);
		mem_tracker_set_addr(ment, gamma);
	}

	return gamma;
}

void *
j_asprintf(const char *format, ...)
{
	va_list ap;
	char *beta = NULL;
	int retval;

	va_start(ap, format);
	retval = vasprintf(&beta, format, ap);
	va_end(ap);
	
	if (!beta || retval <= 0) 
		return NULL;
	
	mem_tracker_add_entry(beta, retval);
	return beta;
}



/*
** Frees an entry and, lala, unlinks it from the junkie memory linked list
*/
void
j_free(void *ptr)
{
	if (!mem_tracker_unlink_addr(ptr)) {
		DEBUG_MSG(DBG_WEIRD, "entry was not allocated with j_free, continuing though");
	}
	
	free(ptr);
	return;

}


/*
** junkie strdup (pronounced strrdup, or shutthefuckup)
*/
char *
j_strdup(const char *str)
{
	char *ptr = NULL;

	if (!str)
		return NULL;
		
	if (!(ptr = j_malloc(strlen(str)+1)))
		return NULL;

	strcpy(ptr, str);
	return ptr;
}
