/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: sound.c,v 1.36 2003/03/03 03:05:34 thepin Exp $
 */

/*
** Play using ESD lib, otherwise fall back to speaker (/dev/speaker).
** Consider whether the ESD should be needed while compiling or if it should
** be checked if it is on the system at runtime.
*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <configure.h>	/* file generated by configure. */
/* local includes */
#include <defaults.h>
#include <sound.h>
#include <sound_handlers.h>
#include <sound_mixer.h>
#include <config.h>
#include <mem.h>
#include <misc.h>

/* list of default sounds, should be possible to change through preferences. */
struct sound_list_default sounds_default[] = {
	{	SOUND_STARTUP,			"welcome.wav",	 		S_LOAD				},
	{	SOUND_LOGIN,			"login.wav",			S_PRELOAD			},
	{	SOUND_TRANSFER_COMPLETE,	"transfer_complete.wav",	S_PRELOAD			},
	{	SOUND_QUIT,			"quit.wav",			S_LOAD				},
	{   0, NULL, 0}
};


struct sound_list *sounds_list = NULL;


/*
** Add a sound to the list of default sounds.
** Those sounds are mainly used by the notification
** sound system.
*/
void
sound_list_add(char *handler, char *path, int type, int esd_id, int load)
{
	struct sound_list **sp = NULL;

	sp = &sounds_list;
	while (*sp)
		sp = &(*sp)->next;
	
	if ( !(*sp = j_malloc(sizeof(struct sound_list))) ) {
		SOUND_DEBUG("sound bug, j_malloc error\n");
		return;
	}

	memset(*sp, 0, sizeof(struct sound_list));

	if (path)
		(*sp)->path = j_strdup(path);
	
	if (handler)
		(*sp)->handler = j_strdup(handler);

	(*sp)->load = load;
	(*sp)->type = type;
	(*sp)->esd_id = esd_id;
	(*sp)->next = NULL;

	return;
}


/*
** Load the standard sound files up (probably most for the notification sound support).
*/
int
sound_list_load()
{
	int i, eid = 0;
	char buf[GOOD_SIZE + 1];
	struct sound_handler *shand = NULL;

	/* should not maybe only get the files there.. */
	for (i = 0; sounds_default[i].file; i++) {
		snprintf(buf, GOOD_SIZE, DATADIR "/junkie/sound/%s", sounds_default[i].file);

		if ( (shand = sound_handler_find_by_filename(buf)) == NULL) { 
			DEBUG_MSG(DBG_WEIRD, "no handler found for %s", buf);
		} else {
			eid = 0;
			if (sounds_default[i].load == S_PRELOAD && shand->preload) {
				/* preload the file if possible. */
				eid = shand->preload(sounds_default[i].file, buf);
				DEBUG_MSG(DBG_INFO, "preloaded %s as id: %d", buf, eid);
			}

			sound_list_add(shand->handler, buf, sounds_default[i].type, eid, sounds_default[i].load);
		}
	}

	return 1;
}


/*
** Free the list of default sounds.
*/
void
sound_list_free()
{
	struct sound_list *snd = NULL, *snd2 = NULL;

	snd = sounds_list;
	while (snd) {
		snd2 = snd;
		snd = snd->next;
		j_free(snd2);
	}

	return;
}

/*
** Find the sound requested by its type.  A type could for example be a startup sound, or a transfer start sound.
** A type exist for various events.
*/
struct sound_list *
sound_list_find_by_type(int type)
{
	struct sound_list *snd = NULL;
	
	snd = sounds_list;
	while (snd) {
		DEBUG_MSG(DBG_INFO, "snd->type: %d", type);
		if (snd->type == type)
			break;
			
		snd = snd->next;
	}

	return snd;
}



int
sound_init(void)
{
//	if (!config_read_bool("sound")) 
//		return 0;

	/*
	** sounds_list load will load all the sounds and cache them for swift playback.
	*/
	DEBUG_MSG(DBG_INFO, "sound_handler_setup");
	sound_handler_setup();
	DEBUG_MSG(DBG_INFO, "sound_handler_all_init()");
	sound_handler_all_init();
	DEBUG_MSG(DBG_INFO, "sound_list_load()");
	sound_list_load();
	DEBUG_MSG(DBG_INFO, "done.");

	sound_mixer_init("/dev/mixer");
	
	return 0;
}



/* shut down the sound support. */
int
sound_shutdown(void)
{
	/* free all the samples setup, and the linked list also XXX */
	sound_stop();
	sound_handler_all_shutdown();
	
	/* free the list of default sounds.  */
	sound_list_free();
	
	sound_mixer_restore();
	
	SOUND_DEBUG("Sound deinitialized");


	
	return 0;
}

/*
** stop playing a sound.
*/
void
sound_stop()
{
	sound_handler_all_stop();
	return;

}


/*
** play a sound by an event type (event type could fx. be starting up, transfer start).
** this also generally means that the file must be be predefined somewhere.
*/
void
sound_play_event(int type)
{
	struct sound_list *snd = NULL;
	struct sound_handler *shand = NULL;
	
	if (!config_read_bool("sound"))
		return;

	/* find the appropriate sound for this event type. */
	snd = sound_list_find_by_type(type);
	if (!snd) {
		DEBUG_MSG(DBG_WEIRD, "was gonna play sound type %d, but not found", type);
		return;
	}

	if (!snd->handler) {
		DEBUG_MSG(DBG_WEIRD, "no handler registered for this sound");
		return;
	}

	/* find an appropriate handler. */
	shand = sound_handler_find_by_name(snd->handler);
	if (!shand || !shand->play_event) {
		SOUND_DEBUG("Unable to play sound for this event.\n");
		return;
	}

	/* play the wanted sound with the right handler.*/
	shand->play_event(snd);
	
	return;
}


/* Play a given sound specified by a path
** difference from sound_play_event is that the file does not need to be predefined anywhere before
** caling this function.
*/
void
sound_play_path(const char *path, int player, cb callback)
{
	struct sound_handler *shand = NULL;
	
	if (!config_read_bool("sound") && !player)
		return;

	/*
	** Find the appropriate handler (if it exist, and call the specified handler for playing this file.
	*/
	if ( (shand = sound_handler_find_by_filename((char *)path)) == NULL) {
		SOUND_DEBUG("Unable to play %s, no handler found.\n"/*, path*/);
		return;
	}

	if (shand->play_path == NULL) {
		SOUND_DEBUG("Failed to play %s, no player function found\n"/*, path*/);
		return;
	}

	/* play yhe file with the correct handler which was found above. */
	shand->play_path((char *)path, callback);

	return;
}
