/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: sound_handlers.c,v 1.13 2003/03/07 17:48:13 thepin Exp $
 */


#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <configure.h>	/* file generated by configure. */

/* local headers. */
#include <sound.h>
#include <sound_handlers.h>
#ifdef HAVE_ESD
#include <sound_esd.h>
#ifdef HAVE_VORBIS
#include <sound_vorbis.h>
#endif
#ifdef HAVE_LIBJAMP
#include <sound_mpg.h>
#endif
#endif
#ifdef HAVE_LIBMIKMOD
#include <sound_mikmod.h>
#endif
#include <sound_speaker.h>
#include <defaults.h>
#include <mem.h>

struct sound_handler *sound_handlers;

/*
** Add a sound handler to the list of sound handlers.
*/
void
sound_handler_add(const char *handler, int priority, const char *formats, void *startup, void *shutdown,
                  void *preload, void *play_event, void *play_path, void *stop)
{
	struct sound_handler **sp = NULL;

	sp = &sound_handlers;
	while (*sp) {
		DEBUG_MSG(DBG_INFO, "handler called %s", (*sp)->handler);
		sp = &(*sp)->next;
	}
	
	if ( !(*sp = j_malloc(sizeof(struct sound_handler))) ) 
		return;

	(*sp)->handler = (*sp)->formats = NULL;
	
	if (handler)
		(*sp)->handler = j_strdup(handler);

	DEBUG_MSG(DBG_INFO, "handler is %s", (*sp)->handler);

	if (formats) {
		/* We want the string with formats to be lowercase and spaces in front and after the string. */
		(*sp)->formats = j_strdup(formats);
	} else
		(*sp)->formats = NULL;

	(*sp)->startup = startup;
	(*sp)->shutdown = shutdown;
	(*sp)->preload = preload;
	(*sp)->play_event = play_event;
	(*sp)->play_path = play_path;
	(*sp)->stop = stop;
	(*sp)->next = NULL;

	return;
}



/*
** Find which handler should be used for
** files with a certain extension.
**/
struct sound_handler *
sound_handler_find_by_extension(char *ext)
{
	struct sound_handler *shand = NULL;

	if (!ext)
		return NULL;
		
	shand = sound_handlers;
	DEBUG_MSG(DBG_INFO, "shand = %p", shand);
	while (shand) {
		DEBUG_MSG(DBG_INFO, "handler: %s ext: \"%s\" formats: %s",
			shand->handler, ext, shand->formats);
		if (shand->formats && strstr(shand->formats, ext)) {
			DEBUG_MSG(DBG_INFO, "matched, found");
			break;	
		}
			
		shand = shand->next;
	}

	return shand;
}


/*
** Find which handler should be used for a special filename.
** Simply finds the extension of the filename and calls another
** function which determines which handler should be used and
** this function passess the information on.
*/
struct sound_handler *
sound_handler_find_by_filename(char *path)
{
	char *ext = NULL;
	
	if (!path)
		return NULL;
	
	if ( (ext = strrchr(path, '.')) == NULL)
		return NULL;
	DEBUG_MSG(DBG_INFO, "extension is: %s", ext);
	return sound_handler_find_by_extension(ext+1);
}

/*
** Find a handler by its registered name.
*/
struct sound_handler *
sound_handler_find_by_name(char *name)
{
	struct sound_handler *shand = NULL;

	shand = sound_handlers;
	while (shand) {
		if (!strcasecmp(shand->handler, name))
			break;
			
		shand = shand->next;
	}

	return shand;
}


/*
** Startup all loaded handlers.
** ie. call their startup function if they got one.
*/
void
sound_handler_all_init(void)
{
	struct sound_handler *shand = NULL;

	shand = sound_handlers;

	DEBUG_MSG(DBG_INFO, "whiling...");
	while (shand) {
		if (shand->startup)
			shand->startup();
			
		DEBUG_MSG(DBG_INFO, "one...");
		shand = shand->next;
	}

	return;
}

/*
** Shutdown all handlers
** ie. call their shutdown functions, if they got one.
*/
void
sound_handler_all_shutdown(void)
{
	struct sound_handler *shand = NULL, *shand2 = NULL;

	shand = sound_handlers;
	while (shand) {
		if (shand->shutdown)
			shand->shutdown();

		shand2 = shand;
		shand = shand->next;
		j_free(shand2);
	}

	return;
}

/*
** all stop playing sound
*/
void
sound_handler_all_stop(void)
{
	struct sound_handler *shand = NULL;

	shand = sound_handlers;
	while (shand) {
		if (shand->stop)
			shand->stop();

		shand = shand->next;
	}

	return;
}


/*
** Setup the sound handlers.
** If u want to add a new sound handler, u need to add somethign here.
*/
void
sound_handler_setup(void)
{
#ifdef HAVE_ESD
	/* The ESD handler. */
	sound_handler_add("esd", 20, " wav au ", sound_esd_init, sound_esd_shutdown, 
			sound_esd_preload, sound_esd_play_event, sound_esd_play_file, NULL);
	
#ifdef HAVE_VORBIS
	/* The Ogg-Vorbis/ESD handler. */
	sound_handler_add("vorbis", 10, " ogg ", sound_vorbis_init, sound_vorbis_shutdown,
			sound_vorbis_preload, sound_vorbis_play_event, sound_vorbis_play_file, sound_vorbis_stop);
#endif
#ifdef HAVE_LIBJAMP
	/* the mpglib library handler. */
	sound_handler_add("libjamp", 10, " mp3 ", sound_mpg_init, sound_mpg_shutdown,
			NULL, NULL, sound_mpg_play_file, sound_mpg_stop);
#endif
#endif

#ifdef HAVE_LIBMIKMOD
	/* The LibMikmod handlder (which handles esd connections by itself).*/
	sound_handler_add("mikmod", 30, " it xm mod s3m ", sound_mikmod_init, NULL,
			NULL, NULL, sound_mikmod_play_file, sound_mikmod_stop);
#endif

	/* Last, but not least, speaker support. */
	sound_handler_add("speaker", 40, NULL, NULL, NULL, NULL, sound_speaker_play_event, NULL, NULL);

	return;
}
