/* junkie - gui ftp client
 * Copyright (c) 2002-2003, Gunnsteinn Hall, Joel Thomas, Martin Faxer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of the
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 *  HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * $Id: sound_mixer.c,v 1.3 2003/03/03 02:16:34 tgucio Exp $
 */


#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
/* local headers */
#include <sound_mixer.h>
#include <mem.h>
#include <defaults.h>

#define MIXERDEV "/dev/mixer"

const char *sound_device_names[] = SOUND_DEVICE_NAMES;



static mixer *mixerdevice = 0;

mixer* 
sound_mixer_init(char *dev)
{

	if(mixerdevice)	return NULL;
	
	mixerdevice = j_malloc(sizeof(mixer));
	memset(mixerdevice, 0, sizeof(mixer));
	
	
	/* attempt to open the mixer device */
	mixerdevice->fd = open(dev, O_RDWR, 0);

	/* detect open error */
	if(mixerdevice->fd == -1) {
		perror(dev);
		return NULL ;
	}
	
	sound_mixer_initialize(mixerdevice);
	sound_mixer_get_level(mixerdevice);
	
	DEBUG_MSG(DBG_INFO, "level is... %d", mixerdevice->left[0]);
	
	return mixerdevice;
}

int
sound_mixer_initialize(mixer *mix)
{
	/* query the mixer capabilities */

	if(ioctl(mix->fd, SOUND_MIXER_READ_DEVMASK,    &mix->outmask)    == -1) {
		perror("unable to query mixer capabilities");
		return -1;
	}
	if(ioctl(mix->fd, SOUND_MIXER_READ_STEREODEVS, &mix->stereomask) == -1) {
		perror("unable to query mixer stereo capabilities");
		return -1;
	}

	return 0;
}

int
sound_mixer_volume_get()
{
	if(!mixerdevice)	mixerdevice = sound_mixer_init(MIXERDEV);
	return mixerdevice->volume;
}

int
sound_mixer_get_level(mixer *mix)
{
	int counter, level;
	char left, right;

	
	for(counter = 0; counter < SOUND_MIXER_NRDEVICES; counter++) {
	
		if(!mix->outmask & (1 << counter) || 
			strcasecmp("vol", sound_device_names[counter]))	
		continue;
		
		level = sound_mixer_get_level_query(mix, counter);
		if(level == -1)	continue;

		left  = (level & 0xFF);
		right = (level & 0xFF00) >> 8;
		
		DEBUG_MSG(DBG_INFO, "level = %d (%d) (%d)", level, left, right);
		
		if(mix->stereomask & (1 << counter))
			mix->left[counter] = left, mix->right[counter] = right;
		else
			mix->left[counter] = mix->right[counter] = left;
	}

	mix->volume = left;
	return level;
}


int
sound_mixer_get_level_query(mixer *mix, int channel)
{
	int volume;

	if(ioctl(mix->fd, MIXER_READ(channel), &volume) == -1) {
		perror("unable to query mixer settings");
		return -1;
	}

	return volume;
}

void
sound_mixer_volume_set(char left, char right)
{
	int level;
	
	if(left  < 0)   left  = 0;
	if(left  > 100) left  = 100;
	if(right < 0)   right = 0;
	if(right > 100) right = 100;

	level = left + (right << 8);
	DEBUG_MSG(DBG_INFO, "LEVEL ========== %d", level);
	
	if(!mixerdevice)    mixerdevice = sound_mixer_init(MIXERDEV);
	sound_mixer_set_level(mixerdevice, level);
}

int
sound_mixer_set_level(mixer *mix, int level)
{
	int counter;

	for(counter	= 0; counter < SOUND_MIXER_NRDEVICES; counter++) {
		
		if (strcasecmp("vol", sound_device_names[counter]))	continue;
		
		if(ioctl(mix->fd, MIXER_WRITE(counter), &level) == -1) {
			perror("unable to set mixer settings");
			return -1;
		}
		break;
	}

	return level;
}

int
sound_mixer_restore()
{
	int counter, level;
	
	if(!mixerdevice)    return -1;

	
	for(counter = 0; counter < SOUND_MIXER_NRDEVICES; counter++) {
		
		if (strcasecmp("vol", sound_device_names[counter]))	continue;
	
		level = mixerdevice->left[counter] + (mixerdevice->right[counter] << 8);
		if(ioctl(mixerdevice->fd, MIXER_WRITE(counter), &level) == -1) {
			perror("unable to restore mixer settings");
			return -1;
		}
		break;
	}
	
	j_free(mixerdevice);
	mixerdevice = NULL;
	return 0;
}

